# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import contextlib
import math

import time

import mock
import pytest

from hikari import errors
from hikari.impl import buckets
from hikari.impl import rate_limits
from hikari.internal import routes
from hikari.internal import time as hikari_date


class TestRESTBucket:
    @pytest.fixture
    def template(self):
        return routes.Route("GET", "/foo/bar")

    @pytest.fixture
    def compiled_route(self, template):
        return routes.CompiledRoute("/foo/bar", template, "1a2b3c")

    @pytest.mark.parametrize(
        ("name", "expected"), [("spaghetti", False), ("UNKNoWN hash", False), ("UNKNOWN", True), ("UNKNOWN hash", True)]
    )
    def test_is_unknown(self, name, compiled_route, expected):
        bucket = buckets.RESTBucket(name, compiled_route, mock.Mock(), float("inf"))

        assert bucket.is_unknown is expected

    @pytest.mark.asyncio
    async def test_usage_when_unknown(self, compiled_route):
        bucket = buckets.RESTBucket(buckets.UNKNOWN_HASH, compiled_route, mock.Mock(), float("inf"))
        bucket._lock = mock.Mock(acquire=mock.AsyncMock(), locked=mock.Mock(return_value=False))

        async with bucket:
            bucket._lock.acquire.assert_awaited_once_with()
            bucket._lock.release.assert_not_called()

            bucket._lock.locked.return_value = True

        bucket._lock.acquire.assert_awaited_once_with()
        bucket._lock.release.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_usage_when_resolved(self, compiled_route):
        global_ratelimit = mock.Mock(acquire=mock.AsyncMock(), reset_at=None)
        bucket = buckets.RESTBucket("resolved bucket", compiled_route, global_ratelimit, float("inf"))
        bucket.remaining = 10
        bucket._lock = mock.Mock(acquire=mock.AsyncMock(), locked=mock.Mock(return_value=False))

        async with bucket:
            bucket._lock.acquire.assert_not_called()
            bucket._lock.release.assert_not_called()

        bucket._lock.acquire.assert_not_called()
        bucket._lock.release.assert_not_called()

    def test_update_rate_limit_when_no_issues(self, compiled_route):
        bucket = buckets.RESTBucket("updating ratelimit test", compiled_route, mock.Mock(), float("inf"))
        now = time.time()

        bucket.remaining = 1
        bucket.limit = 2
        bucket.increase_at = now
        bucket.period = 2

        bucket.update_rate_limit(0, 2, now, 4)

        assert bucket.remaining == 1
        assert bucket.limit == 2
        assert bucket.increase_at == now
        assert bucket.period == 2

    @pytest.mark.parametrize("limit", (10, 5))
    def test_update_rate_limit_when_limit_changed(self, compiled_route, limit):
        bucket = buckets.RESTBucket("updating ratelimit test", compiled_route, mock.Mock(), float("inf"))
        now = time.time()

        bucket.remaining = 1
        bucket.limit = 2
        bucket.increase_at = now
        bucket.period = 2

        bucket.update_rate_limit(0, limit, now + 20, 4)

        assert bucket.limit == limit

    def test_update_rate_limit_when_period_far_apart(self, compiled_route):
        bucket = buckets.RESTBucket("updating ratelimit test", compiled_route, mock.Mock(), float("inf"))
        now = 12123123

        bucket.remaining = 1
        bucket.limit = 3
        bucket.increase_at = now
        bucket.period = 2
        bucket._out_of_sync = True

        bucket.update_rate_limit(1, 3, 123123124, 4.5)

        assert bucket.remaining == 1
        assert bucket.limit == 3
        assert bucket.increase_at == 123123121.75
        assert bucket.period == 2.25
        assert bucket._out_of_sync is False

    @pytest.mark.asyncio
    async def test_acquire_when_too_long_ratelimit(self, compiled_route):
        bucket = buckets.RESTBucket("spaghetti", compiled_route, mock.Mock(), 60)
        bucket.increase_at = time.time() + 999999999999999999999999999
        bucket._lock = mock.Mock(acquire=mock.AsyncMock())

        with (
            mock.patch.object(buckets.RESTBucket, "is_rate_limited", return_value=True),
            pytest.raises(errors.RateLimitTooLongError),
        ):
            await bucket.acquire()

        bucket._lock.acquire.assert_not_called()
        bucket._lock.release.assert_not_called()

    @pytest.mark.asyncio
    async def test_acquire_when_too_long_global_ratelimit(self, compiled_route):
        global_ratelimit = mock.Mock(reset_at=time.time() + 999999999999999999999999999)

        bucket = buckets.RESTBucket("spaghetti", compiled_route, global_ratelimit, 1)
        bucket._lock = mock.Mock(acquire=mock.AsyncMock())

        with (
            mock.patch.object(rate_limits.WindowedBurstRateLimiter, "acquire") as super_acquire,
            pytest.raises(errors.RateLimitTooLongError),
        ):
            await bucket.acquire()

        bucket._lock.acquire.assert_not_called()
        bucket._lock.release.assert_not_called()
        super_acquire.assert_called_once_with()
        global_ratelimit.acquire.assert_not_called()

    @pytest.mark.asyncio
    async def test_acquire_when_unknown_bucket(self, compiled_route):
        global_ratelimit = mock.Mock(acquire=mock.AsyncMock(), reset_at=None)

        bucket = buckets.RESTBucket("UNKNOWN", compiled_route, global_ratelimit, float("inf"))
        bucket._lock = mock.Mock(acquire=mock.AsyncMock())

        with mock.patch.object(rate_limits.WindowedBurstRateLimiter, "acquire") as super_acquire:
            await bucket.acquire()

        bucket._lock.acquire.assert_awaited_once_with()
        super_acquire.assert_not_called()
        global_ratelimit.acquire.assert_not_called()

    @pytest.mark.asyncio
    async def test_acquire_when_resolved_while_waiting(self, compiled_route):
        async def resolve_bucket():
            nonlocal lock_acquire_called

            bucket.resolve("some real hash", 1, 2, 3, 4)
            lock_acquire_called += 1

        global_ratelimit = mock.Mock(acquire=mock.AsyncMock(), reset_at=None)
        bucket = buckets.RESTBucket(buckets.UNKNOWN_HASH, compiled_route, global_ratelimit, float("inf"))
        bucket._lock = mock.Mock(acquire=resolve_bucket)
        lock_acquire_called = 0

        with mock.patch.object(rate_limits.WindowedBurstRateLimiter, "acquire") as super_acquire:
            await bucket.acquire()

        super_acquire.assert_awaited_once_with()
        assert lock_acquire_called == 1
        global_ratelimit.acquire.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_acquire_when_resolved_bucket(self, compiled_route):
        global_ratelimit = mock.Mock(acquire=mock.AsyncMock(), reset_at=None)
        bucket = buckets.RESTBucket("spaghetti", compiled_route, global_ratelimit, float("inf"))
        bucket._lock = mock.Mock()

        with mock.patch.object(rate_limits.WindowedBurstRateLimiter, "acquire") as super_acquire:
            await bucket.acquire()

        super_acquire.assert_awaited_once_with()
        bucket._lock.acquire.assert_not_called()
        global_ratelimit.acquire.assert_awaited_once_with()

    def test_resolve_when_not_unknown(self, compiled_route):
        bucket = buckets.RESTBucket("spaghetti", compiled_route, mock.Mock(), float("inf"))

        with pytest.raises(RuntimeError, match=r"Cannot resolve known bucket"):
            bucket.resolve("test", 1, 2, 3, 4)

        assert bucket.name == "spaghetti"

    def test_resolve(self, compiled_route):
        bucket = buckets.RESTBucket(buckets.UNKNOWN_HASH, compiled_route, mock.Mock(), float("inf"))

        bucket.resolve("test", 1, 3, 123123124, 4)

        assert bucket.name == "test"
        assert bucket.remaining == 1
        assert bucket.limit == 3
        assert bucket.period == 2
        assert bucket.increase_at == 123123122


class TestRESTBucketManager:
    @pytest.fixture
    def bucket_manager(self):
        manager = buckets.RESTBucketManager(max_rate_limit=float("inf"))
        manager._gc_task = object()

        return manager

    def test_max_rate_limit_property(self, bucket_manager):
        bucket_manager._max_rate_limit = object()

        assert bucket_manager.max_rate_limit is bucket_manager._max_rate_limit

    @pytest.mark.asyncio
    async def test_close(self, bucket_manager):
        class GcTaskMock:
            def __init__(self):
                self._awaited_count = 0
                self.cancel = mock.Mock()

            def __await__(self):
                if False:
                    yield  # Turns it into a generator

                self._awaited_count += 1

            def __call__(self):
                return self

            def assert_awaited_once(self):
                assert self._awaited_count == 1

        buckets_array = [mock.Mock() for _ in range(30)]
        bucket_manager._real_hashes_to_buckets = {f"blah{i}": b for i, b in enumerate(buckets_array)}
        bucket_manager._gc_task = gc_task = GcTaskMock()

        await bucket_manager.close()

        assert bucket_manager._real_hashes_to_buckets == {}

        for i, b in enumerate(buckets_array):
            b.close.assert_called_once(), i

        assert bucket_manager._gc_task is None
        gc_task.cancel.assert_called_once_with()
        gc_task.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_start(self, bucket_manager):
        bucket_manager._gc_task = None

        bucket_manager.start()
        assert bucket_manager._gc_task is not None

        # cancel created task
        bucket_manager._gc_task.cancel()
        try:
            await bucket_manager._gc_task
        except asyncio.CancelledError:
            pass

    @pytest.mark.asyncio
    async def test_start_when_already_started(self, bucket_manager):
        bucket_manager._gc_task = object()

        with pytest.raises(errors.ComponentStateConflictError):
            bucket_manager.start()

    @pytest.mark.asyncio
    async def test_gc_makes_gc_pass(self, bucket_manager):
        class ExitError(Exception): ...

        with mock.patch.object(buckets.RESTBucketManager, "_purge_stale_buckets") as purge_stale_buckets:
            with mock.patch.object(asyncio, "sleep", side_effect=[None, ExitError]):
                with pytest.raises(ExitError):
                    await bucket_manager._gc(0.001, 33)

        purge_stale_buckets.assert_called_with(33)

    @pytest.mark.asyncio
    async def test_purge_stale_buckets_any_buckets_that_are_empty_but_still_rate_limited_are_kept_alive(
        self, bucket_manager
    ):
        bucket = mock.Mock()
        bucket.is_empty = True
        bucket.is_unknown = False
        bucket.increase_at = time.time() + 999999999999999999999999999

        bucket_manager._real_hashes_to_buckets["foobar"] = bucket

        bucket_manager._purge_stale_buckets(0)

        assert "foobar" in bucket_manager._real_hashes_to_buckets
        bucket.close.assert_not_called()

    @pytest.mark.asyncio
    async def test_purge_stale_buckets_any_buckets_that_are_empty_but_not_rate_limited_and_not_expired_are_kept_alive(
        self, bucket_manager
    ):
        bucket = mock.Mock()
        bucket.is_empty = True
        bucket.is_unknown = False
        bucket.increase_at = time.time()

        bucket_manager._real_hashes_to_buckets["foobar"] = bucket

        bucket_manager._purge_stale_buckets(10)

        assert "foobar" in bucket_manager._real_hashes_to_buckets
        bucket.close.assert_not_called()

    @pytest.mark.asyncio
    async def test_purge_stale_buckets_any_buckets_that_are_empty_but_not_rate_limited_and_expired_are_closed(
        self, bucket_manager
    ):
        bucket = mock.Mock()
        bucket.is_empty = True
        bucket.is_unknown = False
        bucket.increase_at = time.time() - 999999999999999999999999999

        bucket_manager._real_hashes_to_buckets["foobar"] = bucket

        bucket_manager._purge_stale_buckets(0)

        assert "foobar" not in bucket_manager._real_hashes_to_buckets
        bucket.close.assert_called_once()

    @pytest.mark.asyncio
    async def test_purge_stale_buckets_any_buckets_that_are_not_empty_are_kept_alive(self, bucket_manager):
        bucket = mock.Mock()
        bucket.is_empty = False
        bucket.is_unknown = True
        bucket.increase_at = time.time()

        bucket_manager._real_hashes_to_buckets["foobar"] = bucket

        bucket_manager._purge_stale_buckets(0)

        assert "foobar" in bucket_manager._real_hashes_to_buckets
        bucket.close.assert_not_called()

    @pytest.mark.asyncio
    async def test_acquire_route_when_not_in_routes_to_real_hashes_makes_new_bucket_using_initial_hash(
        self, bucket_manager
    ):
        route = mock.Mock()

        with mock.patch.object(buckets, "_create_authentication_hash", return_value="auth_hash"):
            with mock.patch.object(
                buckets, "_create_unknown_hash", return_value="UNKNOWN;auth_hash;bobs"
            ) as create_unknown_hash:
                bucket_manager.acquire_bucket(route, "auth")

        assert "UNKNOWN;auth_hash;bobs" in bucket_manager._real_hashes_to_buckets
        assert isinstance(bucket_manager._real_hashes_to_buckets["UNKNOWN;auth_hash;bobs"], buckets.RESTBucket)
        create_unknown_hash.assert_called_once_with(route, "auth_hash")

    @pytest.mark.asyncio
    async def test_acquire_route_when_not_in_routes_to_real_hashes_doesnt_cache_route(self, bucket_manager):
        route = mock.Mock()
        route.create_real_bucket_hash = mock.Mock(wraps=lambda initial_hash, auth: initial_hash + ";" + auth + ";bobs")

        bucket_manager.acquire_bucket(route, "auth")

        assert bucket_manager._route_hash_to_bucket_hash.get(hash(route.route)) is None

    @pytest.mark.asyncio
    async def test_acquire_route_when_route_cached_already_obtains_hash_from_route_and_bucket_from_hash(
        self, bucket_manager
    ):
        mock_route_hash = 123123123
        route = mock.Mock()
        route.route = mock.Mock(__hash__=lambda _: mock_route_hash)
        route.create_real_bucket_hash = mock.Mock(return_value="eat pant;1234")
        bucket = mock.Mock(reset_at=time.time() + 999999999999999999999999999)
        bucket_manager._route_hash_to_bucket_hash[mock_route_hash] = "eat pant"
        bucket_manager._real_hashes_to_buckets["eat pant;1234"] = bucket

        assert bucket_manager.acquire_bucket(route, "auth") is bucket

    @pytest.mark.asyncio
    async def test_acquire_route_returns_context_manager(self, bucket_manager):
        route = mock.Mock()

        bucket = mock.Mock(reset_at=time.time() + 999999999999999999999999999)
        with mock.patch.object(buckets, "RESTBucket", return_value=bucket):
            route.create_real_bucket_hash = mock.Mock(
                wraps=lambda initial_hash, auth: initial_hash + ";" + auth + ";bobs"
            )

            assert bucket_manager.acquire_bucket(route, "auth") is bucket

    @pytest.mark.asyncio
    async def test_acquire_unknown_route_returns_context_manager_for_new_bucket(self, bucket_manager):
        mock_route_hash = 123123123
        route = mock.Mock()
        route.route = mock.Mock(__hash__=lambda _: mock_route_hash)
        route.create_real_bucket_hash = mock.Mock(return_value="eat pant;bobs")
        bucket = mock.Mock(reset_at=time.time() + 999999999999999999999999999)
        bucket_manager._route_hash_to_bucket_hash[mock_route_hash] = "eat pant"
        bucket_manager._real_hashes_to_buckets["eat pant;bobs"] = bucket

        assert bucket_manager.acquire_bucket(route, "auth") is bucket

    @pytest.mark.asyncio
    async def test_update_rate_limits_if_wrong_bucket_hash_reroutes_route(self, bucket_manager):
        mock_route_hash = 123123123
        route = mock.Mock()
        route.route = mock.Mock(__hash__=lambda _: mock_route_hash)
        route.create_real_bucket_hash = mock.Mock(wraps=lambda initial_hash, auth: initial_hash + ";" + auth + ";bobs")
        bucket_manager._route_hash_to_bucket_hash[mock_route_hash] = "123"

        with (
            mock.patch.object(buckets, "_create_authentication_hash", return_value="auth_hash"),
            mock.patch.object(buckets, "RESTBucket") as bucket,
        ):
            bucket_manager.update_rate_limits(route, "auth", "blep", 22, 23, 123123.56, 3.56)

        assert bucket_manager._route_hash_to_bucket_hash[mock_route_hash] == "blep"
        assert bucket_manager._real_hashes_to_buckets["blep;auth_hash;bobs"] is bucket.return_value
        bucket.return_value.update_rate_limit.assert_called_once_with(22, 23, 123123.56, 3.56)

    @pytest.mark.asyncio
    async def test_update_rate_limits_if_unknown_bucket_hash_reroutes_route(self, bucket_manager):
        mock_route_hash = 123123123
        route = mock.Mock()
        route.route = mock.Mock(__hash__=lambda _: mock_route_hash)
        route.create_real_bucket_hash = mock.Mock(wraps=lambda initial_hash, auth: initial_hash + ";" + auth + ";bobs")
        bucket_manager._route_hash_to_bucket_hash[mock_route_hash] = "123"
        bucket = mock.Mock()
        bucket_manager._real_hashes_to_buckets["UNKNOWN;auth_hash;bobs"] = bucket

        with (
            mock.patch.object(
                buckets, "_create_authentication_hash", return_value="auth_hash"
            ) as create_authentication_hash,
            mock.patch.object(
                buckets, "_create_unknown_hash", return_value="UNKNOWN;auth_hash;bobs"
            ) as create_unknown_hash,
        ):
            bucket_manager.update_rate_limits(route, "auth", "blep", 22, 23, 123123.53, 3.56)

        assert bucket_manager._route_hash_to_bucket_hash[mock_route_hash] == "blep"
        assert bucket_manager._real_hashes_to_buckets["blep;auth_hash;bobs"] is bucket
        bucket.resolve.assert_called_once_with("blep;auth_hash;bobs", 22, 23, 123123.53, 3.56)
        bucket.update_rate_limit.assert_called_once_with(22, 23, 123123.53, 3.56)
        create_unknown_hash.assert_called_once_with(route, "auth_hash")
        create_authentication_hash.assert_called_once_with("auth")

    @pytest.mark.asyncio
    async def test_update_rate_limits_if_right_bucket_hash_does_nothing_to_hash(self, bucket_manager):
        route = mock.Mock()
        route.create_real_bucket_hash = mock.Mock(wraps=lambda initial_hash, auth: initial_hash + ";" + auth + ";bobs")
        bucket_manager._route_hash_to_bucket_hash[route.route] = "123"
        bucket = mock.Mock(reset_at=time.time() + 999999999999999999999999999)
        bucket_manager._real_hashes_to_buckets["123;auth_hash;bobs"] = bucket

        with mock.patch.object(buckets, "_create_authentication_hash", return_value="auth_hash"):
            bucket_manager.update_rate_limits(route, "auth", "123", 22, 23, 123123.53, 7.65)

        assert bucket_manager._route_hash_to_bucket_hash[route.route] == "123"
        assert bucket_manager._real_hashes_to_buckets["123;auth_hash;bobs"] is bucket
        bucket.update_rate_limit.assert_called_once_with(22, 23, 123123.53, 7.65)

    @pytest.mark.asyncio
    async def test_update_rate_limits_updates_params(self, bucket_manager):
        route = mock.Mock()
        route.create_real_bucket_hash = mock.Mock(wraps=lambda initial_hash, auth: initial_hash + ";" + auth + ";bobs")
        bucket_manager._route_hash_to_bucket_hash[route.route] = "123"
        bucket = mock.Mock(reset_at=time.time() + 999999999999999999999999999)
        bucket_manager._real_hashes_to_buckets["123;auth_hash;bobs"] = bucket

        with mock.patch.object(buckets, "_create_authentication_hash", return_value="auth_hash"):
            bucket_manager.update_rate_limits(route, "auth", "123", 22, 23, 123123123.53, 5.32)

        bucket.update_rate_limit.assert_called_once_with(22, 23, 123123123.53, 5.32)

    @pytest.mark.parametrize(("gc_task", "is_alive"), [(None, False), ("some", True)])
    def test_is_alive(self, bucket_manager, gc_task, is_alive):
        bucket_manager._gc_task = gc_task
        assert bucket_manager.is_alive is is_alive



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Pytest integration."""

from __future__ import annotations

import typing

from pipelines import config
from pipelines import nox

RUN_FLAGS = ["-c", config.PYPROJECT_TOML, "--showlocals"]
COVERAGE_FLAGS = [
    "--cov",
    config.MAIN_PACKAGE,
    "--cov-config",
    config.PYPROJECT_TOML,
    "--cov-report",
    "term",
    "--cov-report",
    f"html:{config.COVERAGE_HTML_PATH}",
    "--cov-report",
    "xml",
]


@nox.session()
def pytest(session: nox.Session) -> None:
    """Run unit tests and measure code coverage.

    Coverage can be disabled with the `--skip-coverage` flag.
    """
    _pytest(session)


@nox.session()
def pytest_all_features(session: nox.Session) -> None:
    """Run unit tests and measure code coverage, using speedup modules.

    Coverage can be disabled with the `--skip-coverage` flag.
    """
    _pytest(session, extras_install=["speedups", "server"], python_flags=("-OO",))


def _pytest(
    session: nox.Session, *, extras_install: typing.Sequence[str] = (), python_flags: typing.Sequence[str] = ()
) -> None:
    nox.sync(session, self=True, extras=extras_install, groups=["pytest"])

    flags = RUN_FLAGS

    if "--coverage" in session.posargs:
        session.posargs.remove("--coverage")
        flags.extend(COVERAGE_FLAGS)

    session.run("python", *python_flags, "-m", "pytest", *flags, *session.posargs, config.TEST_PACKAGE)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime
import operator

import mock
import pytest

from hikari import snowflakes
from hikari.impl import gateway_bot


@pytest.fixture
def raw_id():
    return 537_340_989_808_050_216


@pytest.fixture
def neko_snowflake(raw_id):
    return snowflakes.Snowflake(raw_id)


class TestSnowflake:
    def test_created_at(self, neko_snowflake):
        assert neko_snowflake.created_at == datetime.datetime(
            2019, 1, 22, 18, 41, 15, 283_000, tzinfo=datetime.timezone.utc
        )

    def test_increment(self, neko_snowflake):
        assert neko_snowflake.increment == 40

    def test_internal_process_id(self, neko_snowflake):
        assert neko_snowflake.internal_process_id == 0

    def test_internal_worker_id(self, neko_snowflake):
        assert neko_snowflake.internal_worker_id == 2

    def test_hash(self, neko_snowflake, raw_id):
        assert hash(neko_snowflake) == hash(raw_id)

    def test_index(self, neko_snowflake, raw_id):
        assert operator.index(neko_snowflake) == raw_id

    def test_int_cast(self, neko_snowflake, raw_id):
        assert int(neko_snowflake) == raw_id

    def test_str_cast(self, neko_snowflake, raw_id):
        assert str(neko_snowflake) == str(raw_id)

    def test_repr_cast(self, neko_snowflake, raw_id):
        assert repr(neko_snowflake) == repr(raw_id)

    def test_eq(self, neko_snowflake, raw_id):
        assert neko_snowflake == raw_id
        assert neko_snowflake == snowflakes.Snowflake(raw_id)
        assert str(raw_id) != neko_snowflake

    def test_lt(self, neko_snowflake, raw_id):
        assert neko_snowflake < raw_id + 1

    def test_deserialize(self, neko_snowflake, raw_id):
        assert neko_snowflake == snowflakes.Snowflake(raw_id)

    def test_from_datetime(self):
        result = snowflakes.Snowflake.from_datetime(
            datetime.datetime(2019, 1, 22, 18, 41, 15, 283_000, tzinfo=datetime.timezone.utc)
        )
        assert result == 537340989807788032
        assert isinstance(result, snowflakes.Snowflake)

    def test_min(self):
        assert snowflakes.Snowflake.min() == 0

    def test_max(self):
        assert snowflakes.Snowflake.max() == (1 << 63) - 1


class TestUnique:
    @pytest.fixture
    def neko_unique(self, neko_snowflake):
        class NekoUnique(snowflakes.Unique):
            id = neko_snowflake

        return NekoUnique()

    def test_created_at(self, neko_unique):
        assert neko_unique.created_at == datetime.datetime(
            2019, 1, 22, 18, 41, 15, 283_000, tzinfo=datetime.timezone.utc
        )

    def test_index(self, neko_unique, raw_id):
        assert operator.index(neko_unique) == raw_id

    def test__hash__(self, neko_unique, raw_id):
        assert hash(neko_unique) == hash(raw_id)

    def test__eq__(self, neko_snowflake, raw_id):
        class NekoUnique(snowflakes.Unique):
            id = neko_snowflake

        class NekoUnique2(snowflakes.Unique):
            id = neko_snowflake

        unique1 = NekoUnique()
        unique2 = NekoUnique()

        assert unique1 == unique2
        assert unique1 != NekoUnique2()
        assert unique1 != raw_id


@pytest.mark.parametrize(
    ("guild_id", "expected_id"),
    [(140502780547694592, 2), ("655288690192416778", 1), (snowflakes.Snowflake(105785483455418368), 3)],
)
def test_calculate_shard_id_with_shard_count(guild_id, expected_id):
    assert snowflakes.calculate_shard_id(4, guild_id) == expected_id


@pytest.mark.parametrize(
    ("guild_id", "expected_id"),
    [(140502780547694592, 2), ("115590097100865541", 5), (snowflakes.Snowflake(105785483455418368), 7)],
)
def test_calculate_shard_id_with_app(guild_id, expected_id):
    mock_app = mock.Mock(gateway_bot.GatewayBot, shard_count=8)
    assert snowflakes.calculate_shard_id(mock_app, guild_id) == expected_id



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import typing

import aiohttp
import attrs
import mock
import multidict
import pytest

from hikari import snowflakes
from hikari import undefined
from hikari.internal import data_binding


@attrs.define()
class MyUnique(snowflakes.Unique):
    id: snowflakes.Snowflake = attrs.field(converter=snowflakes.Snowflake)


class TestURLEncodedFormBuilder:
    @pytest.fixture
    def form_builder(self):
        return data_binding.URLEncodedFormBuilder()

    def test_add_field(self, form_builder):
        class TestBytesPayload:
            def __init__(self, value: bytes) -> None:
                self.inner = value

            def __eq__(self, other: typing.Any) -> bool:
                if not isinstance(other, TestBytesPayload):
                    return False

                return self.inner == other.inner

            def __repr__(self) -> str:
                # Make it easier to debug future errors
                return f"TestBytesPayload({self.inner!r})"

        with mock.patch.object(aiohttp, "BytesPayload", new=TestBytesPayload):
            form_builder.add_field("test_name", "test_data", content_type="mimetype")
            form_builder.add_field("test_name2", b"test_data2", content_type="mimetype2")

        assert form_builder._fields == [
            ("test_name", "test_data", "mimetype"),
            ("test_name2", TestBytesPayload(b"test_data2"), "mimetype2"),
        ]

    def test_add_resource(self, form_builder):
        mock_resource = object()

        form_builder.add_resource("lick", mock_resource)

        assert form_builder._resources == [("lick", mock_resource)]

    @pytest.mark.asyncio
    async def test_build(self, form_builder):
        resource1 = mock.Mock()
        resource2 = mock.Mock()
        stream1 = mock.Mock(filename="testing1", mimetype="text")
        stream2 = mock.Mock(filename="testing2", mimetype=None)
        data1 = aiohttp.BytesPayload(b"data1")
        data2 = aiohttp.BytesPayload(b"data2")
        mock_stack = mock.AsyncMock(enter_async_context=mock.AsyncMock(side_effect=[stream1, stream2]))
        executor = object()
        form_builder._fields = [("test_name", data1, "mimetype"), ("test_name2", data2, "mimetype2")]
        form_builder._resources = [("aye", resource1), ("lmao", resource2)]

        with mock.patch.object(aiohttp, "FormData") as mock_form_class:
            assert await form_builder.build(mock_stack, executor) is mock_form_class.return_value

        resource1.stream.assert_called_once_with(executor=executor)
        resource2.stream.assert_called_once_with(executor=executor)
        mock_stack.enter_async_context.assert_has_awaits(
            [mock.call(resource1.stream.return_value), mock.call(resource2.stream.return_value)]
        )
        mock_form_class.return_value.add_field.assert_has_calls(
            [
                mock.call("test_name", data1, content_type="mimetype"),
                mock.call("test_name2", data2, content_type="mimetype2"),
                mock.call("aye", stream1, filename="testing1", content_type="text"),
                mock.call("lmao", stream2, filename="testing2", content_type="application/octet-stream"),
            ]
        )


class TestStringMapBuilder:
    def test_is_mapping(self):
        assert isinstance(data_binding.StringMapBuilder(), typing.Mapping)

    def test_is_multidict(self):
        assert isinstance(data_binding.StringMapBuilder(), multidict.MultiDict)

    def test_starts_empty(self):
        mapping = data_binding.StringMapBuilder()
        assert mapping == {}

    def test_put_undefined(self):
        mapping = data_binding.StringMapBuilder()
        mapping.put("foo", undefined.UNDEFINED)
        assert dict(mapping) == {}

    def test_put_general_value_casts_to_str(self):
        m = mock.Mock()
        mapping = data_binding.StringMapBuilder()
        mapping.put("foo", m)
        assert dict(mapping) == {"foo": m.__str__()}

    def test_duplicate_puts_stores_values_as_sequence(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        mapping = data_binding.StringMapBuilder()
        mapping.put("foo", m1)
        mapping.put("foo", m2)
        assert mapping.getall("foo") == [m1.__str__(), m2.__str__()]
        assert list(mapping.keys()) == ["foo", "foo"]

    def test_put_Unique(self):
        mapping = data_binding.StringMapBuilder()

        mapping.put("myunique", MyUnique(123))
        assert dict(mapping) == {"myunique": "123"}

    def test_put_int(self):
        mapping = data_binding.StringMapBuilder()
        mapping.put("yeet", 420_69)
        assert dict(mapping) == {"yeet": "42069"}

    @pytest.mark.parametrize(
        ("name", "input_val", "expect"), [("a", True, "true"), ("b", False, "false"), ("c", None, "null")]
    )
    def test_put_py_singleton(self, name, input_val, expect):
        mapping = data_binding.StringMapBuilder()
        mapping.put(name, input_val)
        assert dict(mapping) == {name: expect}

    def test_put_with_conversion_uses_return_value(self):
        def convert(_):
            return "yeah, i got called"

        mapping = data_binding.StringMapBuilder()
        mapping.put("blep", "meow", conversion=convert)
        assert dict(mapping) == {"blep": "yeah, i got called"}

    def test_put_with_conversion_passes_raw_input_to_converter(self):
        mapping = data_binding.StringMapBuilder()
        convert = mock.Mock()

        expect = object()
        mapping.put("yaskjgakljglak", expect, conversion=convert)
        convert.assert_called_once_with(expect)

    def test_put_py_singleton_conversion_runs_before_check(self):
        def convert(_):
            return True

        mapping = data_binding.StringMapBuilder()
        mapping.put("im hungry", "yo", conversion=convert)
        assert dict(mapping) == {"im hungry": "true"}


class TestJSONObjectBuilder:
    def test_is_mapping(self):
        assert isinstance(data_binding.JSONObjectBuilder(), typing.Mapping)

    def test_starts_empty(self):
        assert data_binding.JSONObjectBuilder() == {}

    def test_put_undefined(self):
        builder = data_binding.JSONObjectBuilder()
        builder.put("foo", undefined.UNDEFINED)
        assert builder == {}

    def test_put_defined(self):
        m = mock.Mock()
        builder = data_binding.JSONObjectBuilder()
        builder.put("bar", m)
        assert builder == {"bar": m}

    def test_put_with_conversion_uses_conversion_result(self):
        m = mock.Mock()
        convert = mock.Mock()
        builder = data_binding.JSONObjectBuilder()
        builder.put("rawr", m, conversion=convert)
        assert builder == {"rawr": convert()}

    def test_put_with_conversion_passes_raw_input_to_converter(self):
        m = mock.Mock()
        convert = mock.Mock()
        builder = data_binding.JSONObjectBuilder()
        builder.put("bar", m, conversion=convert)
        convert.assert_called_once_with(m)

    def test_put_undefined_and_conversion(self):
        mock_convert = mock.Mock()
        builder = data_binding.JSONObjectBuilder()
        builder.put("foo", undefined.UNDEFINED, conversion=mock_convert)
        assert builder == {}
        mock_convert.assert_not_called()

    def test_put_none_and_conversion(self):
        mock_convert = mock.Mock()
        builder = data_binding.JSONObjectBuilder()
        builder.put("foo", None, conversion=mock_convert)
        assert builder == {"foo": None}
        mock_convert.assert_not_called()

    def test_put_array_undefined(self):
        builder = data_binding.JSONObjectBuilder()
        builder.put_array("dd", undefined.UNDEFINED)
        assert builder == {}

    def test__put_array_defined(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        m3 = mock.Mock()
        builder = data_binding.JSONObjectBuilder()
        builder.put_array("ttt", [m1, m2, m3])
        assert builder == {"ttt": [m1, m2, m3]}

    def test_put_array_with_conversion_uses_conversion_result(self):
        r1 = mock.Mock()
        r2 = mock.Mock()
        r3 = mock.Mock()

        convert = mock.Mock(side_effect=[r1, r2, r3])
        builder = data_binding.JSONObjectBuilder()
        builder.put_array("www", [object(), object(), object()], conversion=convert)
        assert builder == {"www": [r1, r2, r3]}

    def test_put_array_with_conversion_passes_raw_input_to_converter(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        m3 = mock.Mock()

        convert = mock.Mock()
        builder = data_binding.JSONObjectBuilder()
        builder.put_array("xxx", [m1, m2, m3], conversion=convert)
        assert convert.call_args_list[0] == mock.call(m1)
        assert convert.call_args_list[1] == mock.call(m2)
        assert convert.call_args_list[2] == mock.call(m3)

    def test_put_snowflake_undefined(self):
        builder = data_binding.JSONObjectBuilder()
        builder.put_snowflake("nya!", undefined.UNDEFINED)
        assert builder == {}

    @pytest.mark.parametrize(
        ("input_value", "expected_str"),
        [
            (100123, "100123"),
            ("100124", "100124"),
            (MyUnique(100127), "100127"),
            (MyUnique("100129"), "100129"),
            (snowflakes.Snowflake(100125), "100125"),
            (snowflakes.Snowflake("100126"), "100126"),
        ],
    )
    def test_put_snowflake(self, input_value, expected_str):
        builder = data_binding.JSONObjectBuilder()
        builder.put_snowflake("WAWAWA!", input_value)
        assert builder == {"WAWAWA!": expected_str}

    def test_put_snowflake_none(self):
        builder = data_binding.JSONObjectBuilder()
        builder.put_snowflake("wawawa osuremono", None)
        assert builder == {"wawawa osuremono": None}

    @pytest.mark.parametrize(
        ("input_value", "expected_str"),
        [
            (100123, "100123"),
            ("100124", "100124"),
            (MyUnique(100127), "100127"),
            (MyUnique("100129"), "100129"),
            (snowflakes.Snowflake(100125), "100125"),
            (snowflakes.Snowflake("100126"), "100126"),
        ],
    )
    def test_put_snowflake_array_conversions(self, input_value, expected_str):
        builder = data_binding.JSONObjectBuilder()
        builder.put_snowflake_array("WAWAWAH!", [input_value] * 5)
        assert builder == {"WAWAWAH!": [expected_str] * 5}

    def test_put_snowflake_array(self):
        builder = data_binding.JSONObjectBuilder()
        builder.put_snowflake_array("DESU!", [123, 456, 987, 115])
        assert builder == {"DESU!": ["123", "456", "987", "115"]}

    def test_put_snowflake_array_undefined(self):
        builder = data_binding.JSONObjectBuilder()
        builder.put_snowflake_array("test", undefined.UNDEFINED)
        assert builder == {}



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Events related to polls."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("PollVoteCreateEvent", "PollVoteDeleteEvent")

import typing

import attrs

from hikari.events import shard_events
from hikari.internal import attrs_extensions

if typing.TYPE_CHECKING:
    from hikari import snowflakes
    from hikari import traits
    from hikari.api import shard as gateway_shard


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class BasePollVoteEvent(shard_events.ShardEvent):
    """Event base for any event that involves a user voting on a poll."""

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.

    shard: gateway_shard.GatewayShard = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from ShardEvent>>.

    user_id: snowflakes.Snowflake = attrs.field()
    """ID of the user that added their vote to the poll."""

    channel_id: snowflakes.Snowflake = attrs.field()
    """ID of the channel that the poll is in."""

    message_id: snowflakes.Snowflake = attrs.field()
    """ID of the message that the poll is in."""

    guild_id: snowflakes.Snowflake | None = attrs.field()
    """ID of the guild that the poll is in."""

    answer_id: int = attrs.field()
    """ID of the answer that the user voted for."""


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class PollVoteCreateEvent(BasePollVoteEvent):
    """Event that is fired when a user add their vote to a poll.

    If the poll allows multiple selection, one event will be fired for each vote.
    """


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class PollVoteDeleteEvent(BasePollVoteEvent):
    """Event that is fired when a user remove their vote to a poll.

    If the poll allows multiple selection, one event will be fired for each vote.
    """



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import invites
from hikari import urls
from hikari.internal import routes
from tests.hikari import hikari_test_helpers


class TestInviteCode:
    def test_str_operator(self):
        mock_invite = hikari_test_helpers.mock_class_namespace(
            invites.InviteCode, code=mock.PropertyMock(return_value="hikari")
        )()
        assert str(mock_invite) == "https://discord.gg/hikari"


class TestInviteGuild:
    @pytest.fixture
    def model(self):
        return invites.InviteGuild(
            app=mock.Mock(),
            id=123321,
            icon_hash="hi",
            name="bye",
            features=[],
            splash_hash="4o4o4o",
            banner_hash="fofoof",
            description="fkkfkf",
            verification_level=1,
            vanity_url_code=None,
            welcome_screen=None,
            nsfw_level=2,
        )

    def test_make_splash_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_splash_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123321, hash="4o4o4o", size=4096, file_format="JPEG", lossless=True
        )

    def test_splash_url(self, model: invites.InviteGuild):
        splash = object()

        with mock.patch.object(invites.InviteGuild, "make_splash_url", return_value=splash):
            assert model.splash_url is splash

    def test_make_splash_url_when_hash(self, model: invites.InviteGuild):
        model.splash_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_splash_url(ext="url", size=2) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123321, hash="18dnf8dfbakfdh", size=2, file_format="URL", lossless=True
        )

    def test_make_splash_url_when_no_hash(self, model: invites.InviteGuild):
        model.splash_hash = None
        assert model.make_splash_url(ext="png", size=1024) is None

    def test_make_banner_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123321, hash="fofoof", size=4096, file_format="JPEG", lossless=True
        )

    def test_banner_url(self, model: invites.InviteGuild):
        banner = object()

        with mock.patch.object(invites.InviteGuild, "make_banner_url", return_value=banner):
            assert model.banner_url is banner

    def test_make_banner_url_when_hash(self, model: invites.InviteGuild):
        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(ext="url", size=512) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123321, hash="fofoof", size=512, file_format="URL", lossless=True
        )

    def test_make_banner_url_when_format_is_None_and_banner_hash_is_for_gif(self, model: invites.InviteGuild):
        model.banner_hash = "a_18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(ext=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=model.id, hash="a_18dnf8dfbakfdh", size=4096, file_format="GIF", lossless=True
        )

    def test_make_banner_url_when_format_is_None_and_banner_hash_is_not_for_gif(self, model: invites.InviteGuild):
        model.banner_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(ext=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=model.id, hash=model.banner_hash, size=4096, file_format="PNG", lossless=True
        )

    def test_make_banner_url_when_no_hash(self, model: invites.InviteGuild):
        model.banner_hash = None
        assert model.make_banner_url(ext="png", size=2048) is None


class TestInviteWithMetadata:
    def test_uses_left(self):
        mock_invite = hikari_test_helpers.mock_class_namespace(
            invites.InviteWithMetadata, init_=False, max_uses=123, uses=55
        )()

        assert mock_invite.uses_left == 68

    def test_uses_left_when_none(self):
        mock_invite = hikari_test_helpers.mock_class_namespace(invites.InviteWithMetadata, init_=False, max_uses=None)()

        assert mock_invite.uses_left is None



# Hikari Examples - A collection of examples for Hikari.
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""A simple bot that can send an audio file as a voice message."""

from __future__ import annotations

import base64
import os

import numpy as np
from pydub import AudioSegment

import hikari

bot = hikari.GatewayBot(token=os.environ["BOT_TOKEN"])


# This function works at time of writing this (2025.03.22) but
# Discord states that this is an implementation detail and might
# change without notice. You have been warned!
def calculate_waveform(audio_file_path: str) -> tuple[str, float]:
    """Calculate the waveform and the duration from an audio file."""
    # Author note: This function is a python port from here:
    # https://github.com/Vendicated/Vencord/blob/b3bff83dd5040950c55e09bed9e47a60490f81d8/src/plugins/voiceMessages/index.tsx#L145

    # Get an array of samples from the audio file.
    # Note:
    #   To use any other file format then .wav you need
    #   to install ffmpeg or libav
    audio = AudioSegment.from_file(audio_file_path)
    samples = np.array(audio.get_array_of_samples(), dtype=np.float32)

    # Normalize the Samples from -1 to 1
    samples /= np.iinfo(audio.array_type).max

    # Determine the duration of the audio file in seconds
    duration = len(audio) / 1000.0

    # Determine the number of "bins"
    num_bins = np.clip(int(duration * 10), min(32, len(samples)), 256)

    # Determine the amount of samples for each bin
    samples_per_bin = len(samples) // num_bins

    # Create an array for the bins, each bin being 0 at first
    bins = np.zeros(num_bins, dtype=np.uint8)

    for i in range(num_bins):
        start = i * samples_per_bin
        end = start + samples_per_bin
        # Compute root-mean-square for each bin to determine the
        # "volume" of each bin
        rms = np.sqrt(np.mean(np.square(samples[start:end])))
        # Multiply the rms by 255 since each bin has 8 bits
        bins[i] = int(rms * 255)

    # Normalize bins
    max_bin = np.max(bins)
    if max_bin > 0:
        ratio = 1 + (255 / max_bin - 1) * min(1, 100 * (max_bin / 255) ** 3)
        bins = np.minimum(255, (bins * ratio).astype(np.uint8))

    # encode it using base64 since discord wants that
    waveform_encoded = base64.b64encode(bins.tobytes()).decode("utf-8")

    return waveform_encoded, duration


@bot.listen()
async def register_commands(_: hikari.StartingEvent) -> None:
    """Register /audio command."""
    application = await bot.rest.fetch_application()

    commands = [bot.rest.slash_command_builder("audio", "Receive a voice message from the bot!")]

    await bot.rest.set_application_commands(application=application.id, commands=commands)


@bot.listen()
async def handle_interactions(event: hikari.CommandInteractionCreateEvent) -> None:
    """Listen for command interaction."""
    if event.interaction.command_name == "audio":
        waveform, duration = calculate_waveform("./sample.wav")
        await event.app.rest.create_interaction_voice_message_response(
            interaction=event.interaction,
            token=event.interaction.token,
            attachment=hikari.File("./sample.wav"),
            waveform=waveform,
            duration=duration,
        )


bot.run()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import voices
from hikari.events import voice_events


class TestVoiceStateUpdateEvent:
    @pytest.fixture
    def event(self):
        return voice_events.VoiceStateUpdateEvent(
            shard=object(), state=mock.Mock(voices.VoiceState), old_state=mock.Mock(voices.VoiceState)
        )

    def test_app_property(self, event):
        assert event.app is event.state.app

    def test_guild_id_property(self, event):
        event.state.guild_id = 123
        assert event.guild_id == 123

    def test_old_voice_state(self, event):
        event.old_state.guild_id = 123
        assert event.old_state.guild_id == 123


class TestVoiceServerUpdateEvent:
    @pytest.fixture
    def event(self):
        return voice_events.VoiceServerUpdateEvent(
            app=None, shard=object(), guild_id=123, token="token", raw_endpoint="voice.discord.com:123"
        )

    def test_endpoint_property(self, event):
        assert event.endpoint == "wss://voice.discord.com:123"

    def test_endpoint_property_when_raw_endpoint_is_None(self, event):
        event.raw_endpoint = None
        assert event.endpoint is None



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Enum of Discord accepted locales."""

from __future__ import annotations

__all__ = ["Locale"]

import typing

from hikari.internal import enums


@typing.final
class Locale(str, enums.Enum):
    """Possible user/guild locales."""

    ID = "id"
    """Indonesian."""

    DA = "da"
    """Danish."""

    DE = "de"
    """German."""

    EN_GB = "en-GB"
    """English, UK."""

    EN_US = "en-US"
    """English, US."""

    ES_ES = "es-ES"
    """Spanish."""

    FR = "fr"
    """French."""

    HR = "hr"
    """Croatian."""

    IT = "it"
    """Italian."""

    LT = "lt"
    """Lithuanian."""

    HU = "hu"
    """Hungarian."""

    NL = "nl"
    """Dutch."""

    NO = "no"
    """Norwegian."""

    PL = "pl"
    """Polish."""

    PT_BR = "pt-BR"
    """Portuguese, Brazilian."""

    RO = "ro"
    """Romanian."""

    FI = "fi"
    """Finnish."""

    SV_SE = "sv-SE"
    """Swedish."""

    VI = "vi"
    """Vietnamese."""

    TR = "tr"
    """Turkish."""

    CS = "cs"
    """Czech."""

    EL = "el"
    """Greek."""

    BG = "bg"
    """Bulgarian."""

    RU = "ru"
    """Russian."""

    UK = "uk"
    """Ukrainian."""

    HI = "hi"
    """Hindi."""

    TH = "th"
    """Thai."""

    ZH_CN = "zh-CN"
    """Chinese, China."""

    JA = "ja"
    """Japanese."""

    ZH_TW = "zh-TW"
    """Chinese, Taiwan."""

    KO = "ko"
    """Korean."""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import pytest

from hikari import undefined
from hikari.internal import mentions


@pytest.mark.parametrize(
    ("function_input", "expected_output"),
    [
        ((True, True, True, True), {"parse": ["everyone", "users", "roles"], "replied_user": True}),
        ((False, False, False, False), {"parse": []}),
        ((undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED), {"parse": []}),
        ((undefined.UNDEFINED, True, True, True), {"parse": ["roles", "users"], "replied_user": True}),
        ((False, False, [123], [456]), {"parse": [], "users": ["123"], "roles": ["456"]}),
        (
            (True, True, [123, "123", 987], ["213", "456", 456]),
            {"parse": ["everyone"], "users": ["123", "987"], "roles": ["213", "456"], "replied_user": True},
        ),
    ],
)
def test_generate_allowed_mentions(function_input, expected_output):
    returned = mentions.generate_allowed_mentions(*function_input)
    for k, v in expected_output.items():
        if isinstance(v, list):
            returned[k] = sorted(v)

    for k, v in expected_output.items():
        if isinstance(v, list):
            expected_output[k] = sorted(expected_output[k])

    assert returned == expected_output



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Custom data structures used within Hikari's core implementation."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "ExtendedMutableMapping",
    "FreezableDict",
    "KeyT",
    "LimitedCapacityCacheMap",
    "SnowflakeSet",
    "ValueT",
    "get_index_or_slice",
)

import abc
import array
import bisect
import itertools
import sys
import typing

from hikari import snowflakes
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    from typing_extensions import Self

"""Type-hint A type hint used for mapped collection objects."""
KeyT = typing.TypeVar("KeyT", bound=typing.Hashable)
"""Type-hint A type hint used for the type of a mapping's key."""
ValueT = typing.TypeVar("ValueT")
"""Type-hint A type hint used for the type of a mapping's value."""

_LONG_LONG_UNSIGNED: typing.Final[str] = sys.intern("Q")


class ExtendedMutableMapping(typing.MutableMapping[KeyT, ValueT], abc.ABC):
    """The abstract class of mutable mappings used within Hikari.

    These are mutable mappings that have a couple of extra methods to allow
    for further optimised copy operations, as well as the ability to freeze
    the implementation of a mapping to make it read-only.
    """

    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    def copy(self) -> Self:
        """Return a copy of this mapped collection.

        Unlike simply doing [dict(mapping)][], this may rely on internal detail
        around how the data is being stored to allow for a more efficient copy.
        This may look like calling [`dict.copy`][] and wrapping the result in a
        mapped collection.

        !!! note
            Any removal policy on this mapped collection will be copied over.

        Returns
        -------
        MapT[KeyT, ValueT]
            A copy of this mapped collection.
        """

    @abc.abstractmethod
    def freeze(self) -> typing.MutableMapping[KeyT, ValueT]:
        """Return a frozen mapping view of the items in this mapped collection.

        Unlike simply doing [dict(mapping)][], this may rely on internal detail
        around how the data is being stored to allow for a more efficient copy.
        This may look like calling [`dict.copy`][].

        !!! note
            Unlike [`ExtendedMutableMapping.copy`][], this should return a pure
            mapping with no removal policy at all.

        Returns
        -------
        typing.MutableMapping[KeyT, ValueT]
            A frozen mapping view of the items in this mapped collection.
        """


class FreezableDict(ExtendedMutableMapping[KeyT, ValueT]):
    """A mapping that wraps a dict, but can also be frozen."""

    __slots__: typing.Sequence[str] = ("_data",)

    def __init__(self, source: dict[KeyT, ValueT] | None = None, /) -> None:
        self._data = source or {}

    @typing_extensions.override
    def clear(self) -> None:
        self._data.clear()

    @typing_extensions.override
    def copy(self) -> FreezableDict[KeyT, ValueT]:
        return FreezableDict(self._data.copy())

    # TODO: name this something different if it is not physically frozen.
    @typing_extensions.override
    def freeze(self) -> dict[KeyT, ValueT]:
        return self._data.copy()

    @typing_extensions.override
    def __delitem__(self, key: KeyT) -> None:
        del self._data[key]

    @typing_extensions.override
    def __getitem__(self, key: KeyT) -> ValueT:
        return self._data[key]

    @typing_extensions.override
    def __iter__(self) -> typing.Iterator[KeyT]:
        return iter(self._data)

    @typing_extensions.override
    def __len__(self) -> int:
        return len(self._data)

    @typing_extensions.override
    def __setitem__(self, key: KeyT, value: ValueT) -> None:
        self._data[key] = value


class LimitedCapacityCacheMap(ExtendedMutableMapping[KeyT, ValueT]):
    """Implementation of a capacity-limited most-recently-inserted mapping.

    This will start removing the oldest entries after it's maximum capacity is
    reached as new entries are added.

    Parameters
    ----------
    source
        A source dictionary of keys to values to create this from.
    limit
        The limit for how many objects should be stored by this mapping before
        it starts removing the oldest entries.
    on_expire
        A function to call each time an item is garbage collected from this
        map. This should take one positional argument of the same type stored
        in this mapping as the value and should return [`None`][].

        This will always be called after the entry has been removed.
    """

    __slots__: typing.Sequence[str] = ("_data", "_limit", "_on_expire")

    def __init__(
        self,
        source: dict[KeyT, ValueT] | None = None,
        /,
        *,
        limit: int,
        on_expire: typing.Callable[[ValueT], None] | None = None,
    ) -> None:
        self._data: dict[KeyT, ValueT] = source or {}
        self._limit = limit
        self._on_expire = on_expire
        self._garbage_collect()

    @typing_extensions.override
    def clear(self) -> None:
        self._data.clear()

    @typing_extensions.override
    def copy(self) -> LimitedCapacityCacheMap[KeyT, ValueT]:
        return LimitedCapacityCacheMap(self._data.copy(), limit=self._limit, on_expire=self._on_expire)

    @typing_extensions.override
    def freeze(self) -> dict[KeyT, ValueT]:
        return self._data.copy()

    def _garbage_collect(self) -> None:
        while len(self._data) > self._limit:
            value = self._data.pop(next(iter(self._data)))

            if self._on_expire:
                self._on_expire(value)

    @typing_extensions.override
    def __delitem__(self, key: KeyT) -> None:
        del self._data[key]

    @typing_extensions.override
    def __getitem__(self, key: KeyT) -> ValueT:
        return self._data[key]

    @typing_extensions.override
    def __iter__(self) -> typing.Iterator[KeyT]:
        return iter(self._data)

    @typing_extensions.override
    def __len__(self) -> int:
        return len(self._data)

    @typing_extensions.override
    def __setitem__(self, key: KeyT, value: ValueT) -> None:
        self._data[key] = value
        self._garbage_collect()


# TODO: can this be immutable?
class SnowflakeSet(typing.MutableSet[snowflakes.Snowflake]):
    r"""Set of [`hikari.snowflakes.Snowflake`][] objects.

    This internally uses a sorted bisect-array of 64 bit integers to represent
    the information. This reduces the space needed to store these objects
    significantly down to the size of 8 bytes per item.

    In contrast, a regular list would take generally 8 bytes per item just to
    store the memory address, and then a further 28 bytes or more to physically
    store the integral value if it does not get interned by the Python
    implementation you are using. A regular set would consume even more
    space, being a hashtable internally.

    The detail of this implementation has the side effect that searches
    will take $$ \mathcal{O} \left( \log n \right) $$ operations in the worst
    case, and $$ \Omega \left (k \right) $$ in the best case. Average case
    will be $$ \mathcal{O} \left( \log n \right) $$

    Insertions and removals will take $$ \mathcal{O} \left( \log n \right) $$
    operations in the worst case, due to [`bisect`][] using a binary insertion sort
    algorithm internally. Average case will be
    $$ \mathcal{O} \left( \log n \right) $$ and best case will be
    $$ \Omega \left\( k \right) $$

    !!! warning
        This is not thread-safe and must not be iterated across whilst being
        concurrently modified.

    Parameters
    ----------
    *ids
        The IDs to fill this table with.
    """

    __slots__: typing.Sequence[str] = ("_ids",)

    def __init__(self, *ids: int) -> None:
        self._ids: typing.MutableSequence[int] = array.array(_LONG_LONG_UNSIGNED)

        if ids:
            self.add_all(ids)

    @typing_extensions.override
    def add(self, value: int, /) -> None:
        """Add a snowflake to this set."""
        # Always append the first item, as we cannot compare with nothing.
        index = bisect.bisect_left(self._ids, value)
        if index == len(self._ids):
            self._ids.append(value)
        elif self._ids[index] != value:
            self._ids.insert(index, value)

    def add_all(self, sfs: typing.Iterable[int], /) -> None:
        """Add a collection of snowflakes to this set."""
        if not sfs:
            return

        for sf in sfs:
            # Yes, this is repeated code, but we want insertions to be as fast
            # as possible for caching purposes, so reduce the number of function
            # calls as much as possible and reimplement the logic for [`add`][]
            # here.
            index = bisect.bisect_left(self._ids, sf)
            if index == len(self._ids):
                self._ids.append(sf)
            elif self._ids[index] != sf:
                self._ids.insert(index, sf)

    @typing_extensions.override
    def clear(self) -> None:
        """Clear all items from this collection."""
        # Arrays lack a "clear" method.
        self._ids = array.array(_LONG_LONG_UNSIGNED)

    @typing_extensions.override
    def discard(self, value: int, /) -> None:
        """Remove a snowflake from this set if it's present."""
        if not self._ids:
            return

        index = bisect.bisect_left(self._ids, value)

        if index < len(self) and self._ids[index] == value:
            del self._ids[index]

    @typing_extensions.override
    def __contains__(self, value: object) -> bool:
        if not isinstance(value, int):
            return False

        index = bisect.bisect_left(self._ids, value)

        if index < len(self._ids):
            return self._ids[index] == value
        return False

    @typing_extensions.override
    def __iter__(self) -> typing.Iterator[snowflakes.Snowflake]:
        return map(snowflakes.Snowflake, self._ids)

    @typing_extensions.override
    def __len__(self) -> int:
        return len(self._ids)

    @typing_extensions.override
    def __repr__(self) -> str:
        return type(self).__name__ + "(" + ", ".join(map(repr, self._ids)) + ")"

    @typing_extensions.override
    def __sizeof__(self) -> int:
        return super().__sizeof__() + sys.getsizeof(self._ids)

    @typing_extensions.override
    def __str__(self) -> str:
        return "{" + ", ".join(map(repr, self._ids)) + "}"


def get_index_or_slice(
    mapping: typing.Mapping[KeyT, ValueT], index_or_slice: int | slice
) -> ValueT | typing.Sequence[ValueT]:
    """Get a mapping's entry at a given index as if it's a sequence of it's values.

    Parameters
    ----------
    mapping
        The mapping of entries to treat as a sequence.
    index_or_slice
        The index to get an entry to get or slice of multiple entries to get.

    Returns
    -------
    ValueT or typing.Sequence[ValueT]
        The value found at the given integer index or a sequence of the values
        found based on the given slice's rules.

    Raises
    ------
    IndexError
        If [`index_or_slice`][] is an int and is outside the range of the mapping's
        contents.
    """
    if isinstance(index_or_slice, slice):
        return tuple(itertools.islice(mapping.values(), index_or_slice.start, index_or_slice.stop, index_or_slice.step))

    try:
        return next(itertools.islice(mapping.values(), index_or_slice, None))
    except StopIteration:
        raise IndexError(index_or_slice) from None



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import applications
from hikari import monetization
from hikari import snowflakes
from hikari import traits
from hikari.impl import special_endpoints
from hikari.interactions import base_interactions
from hikari.interactions import command_interactions


@pytest.fixture
def mock_app():
    return mock.Mock(traits.CacheAware, rest=mock.AsyncMock())


class TestCommandInteraction:
    @pytest.fixture
    def mock_command_interaction(self, mock_app):
        return command_interactions.CommandInteraction(
            app=mock_app,
            id=snowflakes.Snowflake(2312312),
            type=base_interactions.InteractionType.APPLICATION_COMMAND,
            channel=mock.Mock(id=3123123),
            guild_id=snowflakes.Snowflake(5412231),
            member=object(),
            user=object(),
            token="httptptptptptptptp",
            version=1,
            application_id=snowflakes.Snowflake(43123),
            command_id=snowflakes.Snowflake(3123123),
            command_name="OKOKOK",
            command_type=1,
            options=[],
            resolved=None,
            locale="es-ES",
            guild_locale="en-US",
            app_permissions=543123,
            registered_guild_id=snowflakes.Snowflake(12345678),
            entitlements=[
                monetization.Entitlement(
                    id=snowflakes.Snowflake(123123),
                    sku_id=snowflakes.Snowflake(123123),
                    application_id=snowflakes.Snowflake(123123),
                    guild_id=snowflakes.Snowflake(123123),
                    user_id=snowflakes.Snowflake(123123),
                    type=monetization.EntitlementType.APPLICATION_SUBSCRIPTION,
                    starts_at=None,
                    ends_at=None,
                    is_deleted=False,
                    subscription_id=None,
                )
            ],
            authorizing_integration_owners={
                applications.ApplicationIntegrationType.GUILD_INSTALL: snowflakes.Snowflake(123)
            },
            context=applications.ApplicationContextType.PRIVATE_CHANNEL,
        )

    def test_channel_id_property(self, mock_command_interaction):
        assert mock_command_interaction.channel_id == 3123123

    def test_build_response(self, mock_command_interaction, mock_app):
        mock_app.rest.interaction_message_builder = mock.Mock()
        builder = mock_command_interaction.build_response()

        assert builder is mock_app.rest.interaction_message_builder.return_value
        mock_app.rest.interaction_message_builder.assert_called_once_with(base_interactions.ResponseType.MESSAGE_CREATE)

    def test_build_deferred_response(self, mock_command_interaction, mock_app):
        mock_app.rest.interaction_deferred_builder = mock.Mock()
        builder = mock_command_interaction.build_deferred_response()

        assert builder is mock_app.rest.interaction_deferred_builder.return_value
        mock_app.rest.interaction_deferred_builder.assert_called_once_with(
            base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE
        )


class TestAutocompleteInteraction:
    @pytest.fixture
    def mock_autocomplete_interaction(self, mock_app):
        return command_interactions.AutocompleteInteraction(
            app=mock_app,
            id=snowflakes.Snowflake(2312312),
            type=base_interactions.InteractionType.APPLICATION_COMMAND,
            channel=mock.Mock(3123123),
            guild_id=snowflakes.Snowflake(5412231),
            guild_locale="en-US",
            locale="en-US",
            app_permissions=123321,
            member=object(),
            user=object(),
            token="httptptptptptptptp",
            version=1,
            application_id=snowflakes.Snowflake(43123),
            command_id=snowflakes.Snowflake(3123123),
            command_name="OKOKOK",
            command_type=1,
            options=[],
            registered_guild_id=snowflakes.Snowflake(12345678),
            entitlements=[
                monetization.Entitlement(
                    id=snowflakes.Snowflake(123123),
                    sku_id=snowflakes.Snowflake(123123),
                    application_id=snowflakes.Snowflake(123123),
                    guild_id=snowflakes.Snowflake(123123),
                    user_id=snowflakes.Snowflake(123123),
                    type=monetization.EntitlementType.APPLICATION_SUBSCRIPTION,
                    starts_at=None,
                    ends_at=None,
                    is_deleted=False,
                    subscription_id=None,
                )
            ],
            authorizing_integration_owners={
                applications.ApplicationIntegrationType.GUILD_INSTALL: snowflakes.Snowflake(123)
            },
            context=applications.ApplicationContextType.PRIVATE_CHANNEL,
        )

    @pytest.fixture
    def mock_command_choices(self):
        return [
            special_endpoints.AutocompleteChoiceBuilder(name="a", value="b"),
            special_endpoints.AutocompleteChoiceBuilder(name="foo", value="bar"),
        ]

    def test_build_response(self, mock_autocomplete_interaction, mock_app, mock_command_choices):
        mock_app.rest.interaction_autocomplete_builder = mock.Mock()
        builder = mock_autocomplete_interaction.build_response(mock_command_choices)

        assert builder is mock_app.rest.interaction_autocomplete_builder.return_value
        mock_app.rest.interaction_autocomplete_builder.assert_called_once_with(mock_command_choices)

    @pytest.mark.asyncio
    async def test_create_response(
        self,
        mock_autocomplete_interaction: command_interactions.AutocompleteInteraction,
        mock_app,
        mock_command_choices,
    ):
        await mock_autocomplete_interaction.create_response(mock_command_choices)

        mock_app.rest.create_autocomplete_response.assert_awaited_once_with(
            2312312, "httptptptptptptptp", mock_command_choices
        )



# Hikari Examples - A collection of examples for Hikari.
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""A simple bot with a `!image` command that shows the image for various things.

This demonstrates how many Hikari objects can act as attachments without any
special treatment.
"""

from __future__ import annotations

import os
import re

import hikari

bot = hikari.GatewayBot(token=os.environ["BOT_TOKEN"])


@bot.listen()
async def on_message(event: hikari.GuildMessageCreateEvent) -> None:
    """Listen for messages being created."""
    if not event.is_human or not event.content or not event.content.startswith("!"):
        # Do not respond to bots, webhooks, or messages without content or without a prefix.
        return

    args = event.content[1:].split()

    if args[0] == "image":
        # If args == 1, then we were only provided "image", nothing else
        what = "" if len(args) == 1 else args[1]

        # Since uploading can take some time, we give a visual indicator to the user by typing
        async with bot.rest.trigger_typing(event.channel_id):
            await inspect_image(event, what.lstrip())


async def inspect_image(event: hikari.GuildMessageCreateEvent, what: str) -> None:
    """Inspect the image and respond to the user."""
    # Show the avatar for the given user ID:
    if user_match := re.match(r"<@!?(\d+)>", what):
        user_id = hikari.Snowflake(user_match.group(1))
        user = bot.cache.get_user(user_id) or await bot.rest.fetch_user(user_id)
        await event.message.respond("User avatar", attachment=user.avatar_url or user.default_avatar_url)

    # Show the guild icon:
    elif what.casefold() in {"guild", "server", "here", "this"}:
        guild = event.get_guild()
        if guild is None:
            await event.message.respond("Guild is missing from the cache :(")
            return

        if (icon_url := guild.icon_url) is None:
            await event.message.respond("This guild doesn't have an icon")
        else:
            await event.message.respond("Guild icon", attachment=icon_url)

    # Show the image for the given emoji if there is some content present:
    elif what:
        emoji = hikari.Emoji.parse(what)
        await event.message.respond(emoji.name, attachment=emoji)

    # If nothing was given, we should just return the avatar of the person who ran the command:
    else:
        await event.message.respond(
            "Your avatar", attachment=event.author.avatar_url or event.author.default_avatar_url
        )


bot.run()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Code-style jobs."""

from __future__ import annotations

import pathlib
import shutil
import subprocess
import time

from pipelines import config
from pipelines import nox

GIT = shutil.which("git")


@nox.session()
def reformat_code(session: nox.Session) -> None:
    """Remove trailing whitespace in source and then run ruff code formatter."""
    nox.sync(session, groups=["ruff"])

    remove_trailing_whitespaces(session)

    # At the time of writing, sorting imports is not done when running formatting
    # and needs to be done with ruff check
    # see: https://docs.astral.sh/ruff/formatter/#sorting-imports
    session.run("ruff", "format", *config.PYTHON_REFORMATTING_PATHS)
    session.run("ruff", "check", "--select", "I", "--fix", *config.PYTHON_REFORMATTING_PATHS)


@nox.session(venv_backend="none")
def check_trailing_whitespaces(session: nox.Session) -> None:
    """Check for trailing whitespaces in the project."""
    remove_trailing_whitespaces(session, check_only=True)


def remove_trailing_whitespaces(session: nox.Session, /, *, check_only: bool = False) -> None:
    """Remove trailing whitespaces and ensure LR ends are being used."""
    session.log(f"Searching for stray trailing whitespaces in files ending in {config.REFORMATTING_FILE_EXTS}")

    count = 0
    total = 0

    start = time.perf_counter()
    for raw_path in config.FULL_REFORMATTING_PATHS:
        path = pathlib.Path(raw_path)

        dir_total, dir_count = _remove_trailing_whitespaces_for_directory(
            pathlib.Path(path), session, check_only=check_only
        )

        total += dir_total
        count += dir_count

    end = time.perf_counter()

    remark = "Good job! " if not count else ""
    message = "Had to fix" if not check_only else "Found issues in"
    call = session.error if check_only and count else session.log

    call(
        f"{message} {count} file(s). "
        f"{remark}Took {1_000 * (end - start):.2f}ms to check {total} files in this project."
        + ("\nTry running 'nox -s reformat-code' to fix them" if check_only and count else "")
    )


def _remove_trailing_whitespaces_for_directory(
    root_path: pathlib.Path, session: nox.Session, /, *, check_only: bool
) -> tuple[int, int]:
    total = 0
    count = 0

    for path in root_path.glob("*"):
        if path.is_file():
            if path.name.casefold().endswith(config.REFORMATTING_FILE_EXTS):
                total += 1
                count += _remove_trailing_whitespaces_for_file(path, session, check_only=check_only)
            continue

        dir_total, dir_count = _remove_trailing_whitespaces_for_directory(path, session, check_only=check_only)

        total += dir_total
        count += dir_count

    return total, count


def _remove_trailing_whitespaces_for_file(file: pathlib.Path, session: nox.Session, /, *, check_only: bool) -> bool:
    try:
        lines = file.read_bytes().splitlines(keepends=True)
        new_lines = lines.copy()

        for i in range(len(new_lines)):
            line = lines[i].rstrip(b"\n\r \t")
            line += b"\n"
            new_lines[i] = line

        if lines == new_lines:
            return False

        if check_only:
            session.warn(f"Trailing whitespaces found in {file}")
            return True

        session.log(f"Removing trailing whitespaces present in {file}")

        file.write_bytes(b"".join(lines))

        if GIT is not None:
            result = subprocess.check_call(  # noqa: S603
                [GIT, "add", file, "-vf"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=None
            )
            assert result == 0, f"`git add {file} -v' exited with code {result}"

    except Exception as ex:  # noqa: BLE001
        session.warn("Failed to check", file, "because", type(ex).__name__, ex)

    return True



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Events that are bound to the lifetime of an application.

These are types of hooks that can be invoked on startup or shutdown, which can
be used to initialize other resources, fetch information, and perform checks.
"""

from __future__ import annotations

__all__: typing.Sequence[str] = ("StartedEvent", "StartingEvent", "StoppedEvent", "StoppingEvent")

import typing

import attrs

from hikari.events import base_events
from hikari.internal import attrs_extensions

if typing.TYPE_CHECKING:
    from hikari import traits


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class StartingEvent(base_events.Event):
    """Event that is triggered before the application connects to Discord.

    This will only fire once per startup, so is suitable for
    opening database connections and other resources that need to be
    initialized within a coroutine function.

    !!! warning
        The application will not proceed to connect to Discord until all event
        handlers for this event have completed/terminated. This prevents the
        risk of race conditions occurring (e.g. allowing message events
        to try to access a database that has not yet connected fully).

    If you want to do something _after_ the application has initialized, you
    should consider using [`hikari.events.lifetime_events.StartedEvent`][] instead.
    """

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class StartedEvent(base_events.Event):
    """Event that is triggered after the application has started.

    This will only fire once per startup, so is suitable for
    opening database connections and other resources that need to be
    initialized within a coroutine function.

    If you want to do something _before_ the application connects, you should
    consider using [`hikari.events.lifetime_events.StartingEvent`][] instead.
    """

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class StoppingEvent(base_events.Event):
    """Event that is triggered as soon as the application is requested to close.

    This will fire before the connection is physically disconnected.

    This will only fire once per close, so is suitable for
    closing database connections and other resources that need to be
    closed within a coroutine function.

    !!! warning
        The application will not proceed to disconnect from Discord until all
        event handlers for this event have completed/terminated. This
        prevents the risk of race conditions occurring from code that relies
        on a connection still being available to complete.

    If you want to do something _after_ the disconnection has occurred, you
    should consider using [`hikari.events.lifetime_events.StoppedEvent`][] instead.
    """

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class StoppedEvent(base_events.Event):
    """Event that is triggered once the application has disconnected.

    This will only fire once per close, so is suitable for
    closing database connections and other resources that need to be
    closed within a coroutine function.

    !!! warning
        The application will not proceed to leave the run call until all
        event handlers for this event have completed/terminated. This
        prevents the risk of race conditions occurring where a script may
        terminate the process before a callback can occur.

    If you want to do something when the application is preparing to shut down,
    but _before_ any connection to Discord is closed, you should consider using
    [`hikari.events.lifetime_events.StoppingEvent`][] instead.
    """

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import contextlib
import importlib
import logging
import logging.config
import os
import pathlib
import platform
import string
import sys
import time

import colorlog
import mock
import pytest

from hikari import _about
from hikari.impl import config
from hikari.internal import data_binding
from hikari.internal import net
from hikari.internal import ux
from tests.hikari import hikari_test_helpers


class TestInitLogging:
    def test_when_handlers_already_set_up(self):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[None]))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_basic_config = stack.enter_context(mock.patch.object(logging, "basicConfig"))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))

        with stack:
            ux.init_logging("LOGGING_LEVEL", allow_color=True, force_color=False)

        logging_dict_config.assert_not_called()
        logging_basic_config.assert_not_called()
        colored_formatter.assert_not_called()

    def test_when_handlers_specify_not_to_set_up(self):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[]))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_basic_config = stack.enter_context(mock.patch.object(logging, "basicConfig"))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))

        with stack:
            ux.init_logging(None, allow_color=True, force_color=False)

        logging_dict_config.assert_not_called()
        logging_basic_config.assert_not_called()
        colored_formatter.assert_not_called()

    def test_when_flavour_is_a_dict_and_is_not_incremental(self):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[]))
        stack.enter_context(mock.patch.object(ux, "supports_color", return_value=False))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_file_config = stack.enter_context(mock.patch.object(logging.config, "fileConfig"))
        logging_basic_config = stack.enter_context(mock.patch.object(logging, "basicConfig"))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))

        with stack:
            ux.init_logging({"hikari": {"level": "INFO"}}, allow_color=True, force_color=False)

        logging_file_config.assert_not_called()
        logging_dict_config.assert_called_once_with({"hikari": {"level": "INFO"}})
        logging_basic_config.assert_not_called()
        colored_formatter.assert_not_called()

    def test_when_flavour_is_a_dict_and_is_incremental(self):
        # This will emulate it being populated during the basicConfig call
        def _basicConfig(*args, **kwargs):
            logging_basic_config(*args, **kwargs)
            stack.enter_context(mock.patch.object(logging.root, "handlers", new=[handler]))

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[]))
        stack.enter_context(mock.patch.object(ux, "supports_color", return_value=False))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_file_config = stack.enter_context(mock.patch.object(logging.config, "fileConfig"))
        handler = mock.Mock()
        logging_basic_config = mock.Mock()
        stack.enter_context(mock.patch.object(logging, "basicConfig", new=_basicConfig))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))

        with stack:
            ux.init_logging({"incremental": True, "hikari": {"level": "INFO"}}, allow_color=True, force_color=False)

        logging_file_config.assert_not_called()
        logging_dict_config.assert_called_once_with({"incremental": True, "hikari": {"level": "INFO"}})
        logging_basic_config.assert_called_once_with(
            level=None, stream=sys.stdout, format="%(levelname)-1.1s %(asctime)23.23s %(name)s: %(message)s"
        )
        colored_formatter.assert_not_called()
        handler.setFormatter.assert_not_called()

    def test_when_supports_color(self):
        # This will emulate it being populated during the basicConfig call
        def _basicConfig(*args, **kwargs):
            logging_basic_config(*args, **kwargs)
            stack.enter_context(mock.patch.object(logging.root, "handlers", new=[handler]))

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[]))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_file_config = stack.enter_context(mock.patch.object(logging.config, "fileConfig"))
        handler = mock.Mock()
        logging_basic_config = mock.Mock()
        stack.enter_context(mock.patch.object(logging, "basicConfig", new=_basicConfig))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))
        supports_color = stack.enter_context(mock.patch.object(ux, "supports_color", return_value=True))

        with stack:
            ux.init_logging("LOGGING_LEVEL", allow_color=True, force_color=False)

        logging_file_config.assert_not_called()
        logging_dict_config.assert_not_called()
        logging_basic_config.assert_called_once_with(level="LOGGING_LEVEL", stream=sys.stdout)
        colored_formatter.assert_called_once_with(
            fmt="%(log_color)s%(bold)s%(levelname)-1.1s%(thin)s %(asctime)23.23s %(bold)s%(name)s: "
            "%(thin)s%(message)s%(reset)s",
            force_color=True,
        )
        supports_color.assert_called_once_with(allow_color=True, force_color=False)
        handler.setFormatter.assert_called_once_with(colored_formatter.return_value)

    def test_when_doesnt_support_color(self):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[]))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_basic_config = stack.enter_context(mock.patch.object(logging, "basicConfig"))
        logging_file_config = stack.enter_context(mock.patch.object(logging.config, "fileConfig"))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))
        supports_color = stack.enter_context(mock.patch.object(ux, "supports_color", return_value=False))

        with stack:
            ux.init_logging("LOGGING_LEVEL", allow_color=True, force_color=False)

        logging_dict_config.assert_not_called()
        logging_file_config.assert_not_called()
        logging_basic_config.assert_called_once_with(
            level="LOGGING_LEVEL", format="%(levelname)-1.1s %(asctime)23.23s %(name)s: %(message)s", stream=sys.stdout
        )
        colored_formatter.assert_not_called()
        supports_color.assert_called_once_with(allow_color=True, force_color=False)

    def test_when_flavour_is_pathlike(self):
        path = pathlib.Path("some/path/to/somewhere.ini")

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[]))
        pathlib_path = stack.enter_context(mock.patch.object(pathlib, "Path"))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_basic_config = stack.enter_context(mock.patch.object(logging, "basicConfig"))
        logging_file_config = stack.enter_context(mock.patch.object(logging.config, "fileConfig"))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))
        supports_color = stack.enter_context(mock.patch.object(ux, "supports_color", return_value=False))

        with stack:
            ux.init_logging(path, allow_color=True, force_color=False)

        pathlib_path.assert_not_called()
        logging_dict_config.assert_not_called()
        logging_file_config.assert_called_once_with(path)
        logging_basic_config.assert_not_called()
        colored_formatter.assert_not_called()
        supports_color.assert_not_called()

    def test_when_flavour_is_an_existing_path(self):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(logging, "root", handlers=[]))
        pathlib_path = stack.enter_context(mock.patch.object(pathlib, "Path"))
        logging_dict_config = stack.enter_context(mock.patch.object(logging.config, "dictConfig"))
        logging_basic_config = stack.enter_context(mock.patch.object(logging, "basicConfig"))
        logging_file_config = stack.enter_context(mock.patch.object(logging.config, "fileConfig"))
        colored_formatter = stack.enter_context(mock.patch.object(colorlog.formatter, "ColoredFormatter"))
        supports_color = stack.enter_context(mock.patch.object(ux, "supports_color", return_value=False))

        with stack:
            ux.init_logging("some/path/to/somewhere.ini", allow_color=True, force_color=False)

        pathlib_path.assert_called_once_with("some/path/to/somewhere.ini")
        logging_dict_config.assert_not_called()
        logging_file_config.assert_called_once_with(pathlib_path.return_value)
        logging_basic_config.assert_not_called()
        colored_formatter.assert_not_called()
        supports_color.assert_not_called()


def test_read_banner():
    class MockFile:
        context_entered = 0
        context_exited = 0

        def __enter__(self):
            self.context_entered += 1
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.context_exited += 1
            return None

        def read(self):
            return read

    class MockTraversable:
        joint_path = None
        open_mode = None
        mock_file = None
        open_encoding = None

        def joinpath(self, path):
            self.joint_path = path
            return self

        def open(self, mode, encoding):
            self.open_mode = mode
            self.open_encoding = encoding
            return self.mock_file

    traversable = MockTraversable()
    traversable.mock_file = MockFile()
    read = object()

    with mock.patch.object(importlib.resources, "files", return_value=traversable, create=True) as read_text:
        assert ux._read_banner("hikaru") is read

    read_text.assert_called_once_with("hikaru")
    assert traversable.joint_path == "banner.txt"
    assert traversable.open_mode == "r"
    assert traversable.open_encoding == "utf-8"
    assert traversable.mock_file.context_entered == 1
    assert traversable.mock_file.context_exited == 1


class TestPrintBanner:
    def test_when_package_is_none(self):
        with mock.patch.object(sys.stdout, "write") as write:
            ux.print_banner(None, allow_color=True, force_color=False)

        write.assert_not_called()

    @pytest.fixture
    def mock_args(self):
        _about_path = str(pathlib.Path("/", "somewhere", "..", "coding", "hikari", "_about.py"))

        stack = contextlib.ExitStack()

        with stack:
            stack.enter_context(mock.patch.object(platform, "release", return_value="1.0.0"))
            stack.enter_context(mock.patch.object(platform, "system", return_value="Potato"))
            stack.enter_context(mock.patch.object(platform, "machine", return_value="Machine"))
            stack.enter_context(mock.patch.object(platform, "python_implementation", return_value="CPython"))
            stack.enter_context(mock.patch.object(platform, "python_version", return_value="4.0.0"))

            stack.enter_context(mock.patch.object(_about, "__version__", new="2.2.2"))
            stack.enter_context(mock.patch.object(_about, "__git_sha1__", new="12345678901234567890"))
            stack.enter_context(mock.patch.object(_about, "__copyright__", new="2020, Nekokatt"))
            stack.enter_context(mock.patch.object(_about, "__license__", new="MIT"))
            stack.enter_context(mock.patch.object(_about, "__file__", new=_about_path))
            stack.enter_context(mock.patch.object(_about, "__docs__", new="https://nekokatt.github.io/hikari/docs"))
            stack.enter_context(mock.patch.object(_about, "__discord_invite__", new="https://discord.gg/Jx4cNGG"))
            stack.enter_context(mock.patch.object(_about, "__url__", new="https://nekokatt.github.io/hikari"))

            yield None

    def test_when_supports_color(self, mock_args):
        stack = contextlib.ExitStack()
        stack.enter_context(
            mock.patch.object(colorlog.escape_codes, "escape_codes", new={"red": 0, "green": 1, "blue": 2})
        )
        stack.enter_context(mock.patch.object(time, "sleep"))
        supports_color = stack.enter_context(mock.patch.object(ux, "supports_color", return_value=True))
        read_banner = stack.enter_context(mock.patch.object(ux, "_read_banner"))
        template = stack.enter_context(mock.patch.object(string, "Template"))
        stdout = stack.enter_context(mock.patch.object(sys, "stdout"))

        with stack:
            ux.print_banner("hikaru", allow_color=True, force_color=False)

        args = {
            # Hikari stuff.
            "hikari_version": "2.2.2",
            "hikari_git_sha1": "12345678",
            "hikari_copyright": "2020, Nekokatt",
            "hikari_license": "MIT",
            "hikari_install_location": str(pathlib.Path("/", "coding", "hikari").resolve()),
            "hikari_documentation_url": "https://nekokatt.github.io/hikari/docs",
            "hikari_discord_invite": "https://discord.gg/Jx4cNGG",
            "hikari_source_url": "https://nekokatt.github.io/hikari",
            "python_implementation": "CPython",
            "python_version": "4.0.0",
            "system_description": "Machine Potato 1.0.0",
            "red": 0,
            "green": 1,
            "blue": 2,
        }

        read_banner.assert_called_once_with("hikaru")
        template.assert_called_once_with(read_banner.return_value)
        safe_substitute = template.return_value.safe_substitute
        safe_substitute.assert_called_once_with(args)
        supports_color.assert_called_once_with(allow_color=True, force_color=False)
        stdout.buffer.write.assert_called_once_with(safe_substitute.return_value.encode.return_value)
        stdout.flush.assert_called_once_with()

    def test_when_doesnt_supports_color(self, mock_args):
        stack = contextlib.ExitStack()
        stack.enter_context(
            mock.patch.object(colorlog.escape_codes, "escape_codes", new={"red": 0, "green": 1, "blue": 2})
        )
        stack.enter_context(mock.patch.object(time, "sleep"))
        supports_color = stack.enter_context(mock.patch.object(ux, "supports_color", return_value=False))
        read_banner = stack.enter_context(mock.patch.object(ux, "_read_banner"))
        template = stack.enter_context(mock.patch.object(string, "Template"))
        stdout = stack.enter_context(mock.patch.object(sys, "stdout"))

        with stack:
            ux.print_banner("hikaru", allow_color=True, force_color=False)

        args = {
            # Hikari stuff.
            "hikari_version": "2.2.2",
            "hikari_git_sha1": "12345678",
            "hikari_copyright": "2020, Nekokatt",
            "hikari_license": "MIT",
            "hikari_install_location": str(pathlib.Path("/", "coding", "hikari").resolve()),
            "hikari_documentation_url": "https://nekokatt.github.io/hikari/docs",
            "hikari_discord_invite": "https://discord.gg/Jx4cNGG",
            "hikari_source_url": "https://nekokatt.github.io/hikari",
            "python_implementation": "CPython",
            "python_version": "4.0.0",
            "system_description": "Machine Potato 1.0.0",
            "red": "",
            "green": "",
            "blue": "",
        }

        template.assert_called_once_with(read_banner.return_value)
        safe_substitute = template.return_value.safe_substitute
        safe_substitute.assert_called_once_with(args)
        supports_color.assert_called_once_with(allow_color=True, force_color=False)
        stdout.buffer.write.assert_called_once_with(safe_substitute.return_value.encode.return_value)
        stdout.flush.assert_called_once_with()

    def test_use_extra_args(self, mock_args):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(colorlog.escape_codes, "escape_codes", new={}))
        stack.enter_context(mock.patch.object(time, "sleep"))
        read_banner = stack.enter_context(mock.patch.object(ux, "_read_banner"))
        template = stack.enter_context(mock.patch.object(string, "Template"))
        stdout = stack.enter_context(mock.patch.object(sys, "stdout"))

        extra_args = {"extra_argument_1": "one", "extra_argument_2": "two"}

        with stack:
            ux.print_banner("hikaru", allow_color=True, force_color=False, extra_args=extra_args)

        args = {
            # Hikari stuff.
            "hikari_version": "2.2.2",
            "hikari_git_sha1": "12345678",
            "hikari_copyright": "2020, Nekokatt",
            "hikari_license": "MIT",
            "hikari_install_location": str(pathlib.Path("/", "coding", "hikari").resolve()),
            "hikari_documentation_url": "https://nekokatt.github.io/hikari/docs",
            "hikari_discord_invite": "https://discord.gg/Jx4cNGG",
            "hikari_source_url": "https://nekokatt.github.io/hikari",
            "python_implementation": "CPython",
            "python_version": "4.0.0",
            "system_description": "Machine Potato 1.0.0",
        }

        args.update(extra_args)

        read_banner.assert_called_once_with("hikaru")
        template.assert_called_once_with(read_banner.return_value)
        safe_substitute = template.return_value.safe_substitute
        safe_substitute.assert_called_once_with(args)
        stdout.buffer.write.assert_called_once_with(safe_substitute.return_value.encode.return_value)
        stdout.flush.assert_called_once_with()

    def test_overwrite_args_raises_error(self, mock_args):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(time, "sleep"))
        stack.enter_context(mock.patch.object(colorlog.escape_codes, "escape_codes", new={}))
        stack.enter_context(mock.patch.object(ux, "_read_banner"))
        stack.enter_context(mock.patch.object(string, "Template"))
        stack.enter_context(mock.patch.object(sys.stdout, "write"))
        stack.enter_context(mock.patch.object(os.path, "abspath", return_value="some path"))

        with stack:
            with pytest.raises(
                ValueError, match=r"Cannot overwrite \$-substitution `hikari_version`. Please use a different key."
            ):
                ux.print_banner(
                    "hikari", allow_color=True, force_color=False, extra_args={"hikari_version": "overwrite"}
                )


class TestWarnIfNotOptimized:
    @pytest.mark.skipif(not __debug__, reason="Not running in optimized mode")
    def test_when_not_optimized(self):
        with mock.patch.object(ux, "_LOGGER") as logger:
            ux.warn_if_not_optimized(suppress=False)

        logger.warning.assert_called()

    @pytest.mark.skipif(__debug__, reason="Running in optimized mode")
    def test_when_optimized(self):
        with mock.patch.object(ux, "_LOGGER") as logger:
            ux.warn_if_not_optimized(suppress=False)

        logger.warning.assert_not_called()

    @pytest.mark.skipif(not __debug__, reason="Not running in optimized mode")
    def test_when_optimized_and_suppressed(self):
        with mock.patch.object(ux, "_LOGGER") as logger:
            ux.warn_if_not_optimized(suppress=True)

        logger.warning.assert_not_called()


class TestSupportsColor:
    def test_when_not_allow_color(self):
        assert ux.supports_color(allow_color=False, force_color=True) is False

    def test_when_CLICOLOR_FORCE_in_env(self):
        with mock.patch.dict(os.environ, {"CLICOLOR_FORCE": "1"}, clear=True):
            assert ux.supports_color(allow_color=True, force_color=False) is True

    def test_when_force_color(self):
        with mock.patch.dict(os.environ, {"CLICOLOR_FORCE": "0"}, clear=True):
            assert ux.supports_color(allow_color=True, force_color=True) is True

    def test_when_CLICOLOR_and_is_a_tty(self):
        with mock.patch.object(sys.stdout, "isatty", return_value=True):
            with mock.patch.dict(os.environ, {"CLICOLOR_FORCE": "0", "CLICOLOR": "1"}, clear=True):
                assert ux.supports_color(allow_color=True, force_color=False) is True

    def test_when_CLICOLOR_is_0(self):
        with mock.patch.object(sys.stdout, "isatty", return_value=True):
            with mock.patch.dict(os.environ, {"CLICOLOR_FORCE": "0", "CLICOLOR": "0"}, clear=True):
                assert ux.supports_color(allow_color=True, force_color=False) is False

    @pytest.mark.parametrize("colorterm", ["truecolor", "24bit", "TRUECOLOR", "24BIT"])
    def test_when_COLORTERM_has_correct_value(self, colorterm):
        with mock.patch.dict(os.environ, {"COLORTERM": colorterm}, clear=True):
            assert ux.supports_color(allow_color=True, force_color=False) is True

    def test_when_plat_is_Pocket_PC(self):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.dict(os.environ, {}, clear=True))
        stack.enter_context(mock.patch.object(sys, "platform", new="Pocket PC"))

        with stack:
            assert ux.supports_color(allow_color=True, force_color=False) is False

    @pytest.mark.parametrize(
        ("term_program", "ansicon", "isatty", "expected"),
        [
            ("mintty", False, True, True),
            ("Terminus", False, True, True),
            ("some other", True, True, True),
            ("some other", False, True, False),
            ("some other", False, False, False),
            ("mintty", True, False, False),
            ("Terminus", True, False, False),
        ],
    )
    def test_when_plat_is_win32(self, term_program, ansicon, isatty, expected):
        environ = {"TERM_PROGRAM": term_program}
        if ansicon:
            environ["ANSICON"] = "ooga booga"

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.dict(os.environ, environ, clear=True))
        stack.enter_context(mock.patch.object(sys.stdout, "isatty", return_value=isatty))
        stack.enter_context(mock.patch.object(sys, "platform", new="win32"))

        with stack:
            assert ux.supports_color(allow_color=True, force_color=False) is expected

    @pytest.mark.parametrize("isatty", [True, False])
    def test_when_plat_is_not_win32(self, isatty):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.dict(os.environ, {}, clear=True))
        stack.enter_context(mock.patch.object(sys.stdout, "isatty", return_value=isatty))
        stack.enter_context(mock.patch.object(sys, "platform", new="linux"))

        with stack:
            assert ux.supports_color(allow_color=True, force_color=False) is isatty

    @pytest.mark.parametrize("isatty", [True, False])
    @pytest.mark.parametrize("plat", ["linux", "win32"])
    def test_when_PYCHARM_HOSTED(self, isatty, plat):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.dict(os.environ, {"PYCHARM_HOSTED": "OOGA BOOGA"}, clear=True))
        stack.enter_context(mock.patch.object(sys.stdout, "isatty", return_value=isatty))
        stack.enter_context(mock.patch.object(sys, "platform", new=plat))

        with stack:
            assert ux.supports_color(allow_color=True, force_color=False) is True

    @pytest.mark.parametrize("isatty", [True, False])
    @pytest.mark.parametrize("plat", ["linux", "win32"])
    def test_when_WT_SESSION(self, isatty, plat):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.dict(os.environ, {"WT_SESSION": "OOGA BOOGA"}, clear=True))
        stack.enter_context(mock.patch.object(sys.stdout, "isatty", return_value=isatty))
        stack.enter_context(mock.patch.object(sys, "platform", new=plat))

        with stack:
            assert ux.supports_color(allow_color=True, force_color=False) is True


class TestHikariVersion:
    @pytest.mark.parametrize("v", ["1", "1.0.0dev2"])
    def test_init_when_version_number_is_invalid(self, v):
        with pytest.raises(ValueError, match=rf"Invalid version: '{v}'"):
            ux.HikariVersion(v)

    def test_init_when_prerelease(self):
        assert ux.HikariVersion("1.2.3.dev99").prerelease == (".dev", 99)

    def test_init_when_no_prerelease(self):
        assert ux.HikariVersion("1.2.3").prerelease is None

    def test_str_when_prerelease(self):
        assert str(ux.HikariVersion("1.2.3.dev99")) == "1.2.3.dev99"

    def test_str_when_no_prerelease(self):
        assert str(ux.HikariVersion("1.2.3")) == "1.2.3"

    def test_repr(self):
        assert repr(ux.HikariVersion("1.2.3.dev99")) == "HikariVersion('1.2.3.dev99')"

    @pytest.mark.parametrize(
        ("other", "result"),
        [
            (ux.HikariVersion("1.2.3.dev99"), True),
            (ux.HikariVersion("42.212.4.dev99"), False),
            (ux.HikariVersion("1.2.3.dev98"), False),
            (ux.HikariVersion("1.2.3"), False),
        ],
    )
    def test_eq(self, other, result):
        assert (ux.HikariVersion("1.2.3.dev99") == other) is result

    @pytest.mark.parametrize(
        ("other", "result"),
        [
            (ux.HikariVersion("1.2.3.dev99"), False),
            (ux.HikariVersion("42.212.4.dev99"), True),
            (ux.HikariVersion("1.2.3.dev98"), True),
            (ux.HikariVersion("1.2.3"), True),
        ],
    )
    def test_ne(self, other, result):
        assert (ux.HikariVersion("1.2.3.dev99") != other) is result

    @pytest.mark.parametrize(
        ("other", "result"),
        [
            (ux.HikariVersion("1.2.3.dev99"), False),
            (ux.HikariVersion("42.212.4.dev99"), True),
            (ux.HikariVersion("1.2.3.dev98"), False),
            (ux.HikariVersion("1.2.3"), True),
        ],
    )
    def test_lt(self, other, result):
        assert (ux.HikariVersion("1.2.3.dev99") < other) is result

    @pytest.mark.parametrize(
        ("other", "result"),
        [
            (ux.HikariVersion("1.2.3.dev99"), True),
            (ux.HikariVersion("42.212.4.dev99"), True),
            (ux.HikariVersion("1.2.3.dev98"), False),
            (ux.HikariVersion("1.2.3"), True),
        ],
    )
    def test_le(self, other, result):
        assert (ux.HikariVersion("1.2.3.dev99") <= other) is result

    @pytest.mark.parametrize(
        ("other", "result"),
        [
            (ux.HikariVersion("1.2.3.dev99"), False),
            (ux.HikariVersion("42.212.4.dev99"), False),
            (ux.HikariVersion("1.2.3.dev98"), True),
            (ux.HikariVersion("1.2.3"), False),
        ],
    )
    def test_ge(self, other, result):
        assert (ux.HikariVersion("1.2.3.dev99") > other) is result

    @pytest.mark.parametrize(
        ("other", "result"),
        [
            (ux.HikariVersion("1.2.3.dev99"), True),
            (ux.HikariVersion("42.212.4.dev99"), False),
            (ux.HikariVersion("1.2.3.dev98"), True),
            (ux.HikariVersion("1.2.3"), False),
        ],
    )
    def test_gt(self, other, result):
        assert (ux.HikariVersion("1.2.3.dev99") >= other) is result


@pytest.mark.asyncio
class TestCheckForUpdates:
    @pytest.fixture
    def http_settings(self):
        return mock.Mock(spec_set=config.HTTPSettings)

    @pytest.fixture
    def proxy_settings(self):
        return mock.Mock(spec_set=config.ProxySettings)

    async def test_when_not_official_pypi_release(self, http_settings, proxy_settings):
        stack = contextlib.ExitStack()
        logger = stack.enter_context(mock.patch.object(ux, "_LOGGER"))
        create_client_session = stack.enter_context(mock.patch.object(net, "create_client_session"))
        stack.enter_context(mock.patch.object(_about, "__git_sha1__", new="HEAD"))

        with stack:
            await ux.check_for_updates(http_settings=http_settings, proxy_settings=proxy_settings)

        logger.warning.assert_not_called()
        logger.info.assert_not_called()
        create_client_session.assert_not_called()

    async def test_when_error_fetching(self, http_settings, proxy_settings):
        ex = RuntimeError("testing")
        stack = contextlib.ExitStack()
        logger = stack.enter_context(mock.patch.object(ux, "_LOGGER"))
        stack.enter_context(mock.patch.object(_about, "__git_sha1__", new="1234567890"))
        create_client_session = stack.enter_context(mock.patch.object(net, "create_client_session", side_effect=ex))
        create_tcp_connector = stack.enter_context(mock.patch.object(net, "create_tcp_connector"))

        with stack:
            await ux.check_for_updates(http_settings=http_settings, proxy_settings=proxy_settings)

        logger.warning.assert_called_once_with("Failed to fetch hikari version details", exc_info=ex)
        create_tcp_connector.assert_called_once_with(dns_cache=False, http_settings=http_settings)
        create_client_session.assert_called_once_with(
            connector=create_tcp_connector(),
            connector_owner=True,
            http_settings=http_settings,
            raise_for_status=True,
            trust_env=proxy_settings.trust_env,
        )

    async def test_when_no_new_available_releases(self, http_settings, proxy_settings):
        data = {
            "releases": {
                "0.1.0": [{"yanked": False}],
                "1.0.0": [{"yanked": False}],
                "1.0.0.dev1": [{"yanked": False}],
                "1.0.1": [{"yanked": True}],
            }
        }
        _request = hikari_test_helpers.AsyncContextManagerMock()
        _request.read = mock.AsyncMock(return_value=data)
        _client_session = hikari_test_helpers.AsyncContextManagerMock()
        _client_session.get = mock.Mock(return_value=_request)
        stack = contextlib.ExitStack()
        logger = stack.enter_context(mock.patch.object(ux, "_LOGGER"))
        json_loads = stack.enter_context(mock.patch.object(data_binding, "default_json_loads", return_value=data))
        create_client_session = stack.enter_context(
            mock.patch.object(net, "create_client_session", return_value=_client_session)
        )
        create_tcp_connector = stack.enter_context(mock.patch.object(net, "create_tcp_connector"))
        stack.enter_context(mock.patch.object(_about, "__version__", new="1.0.0"))
        stack.enter_context(mock.patch.object(_about, "__git_sha1__", new="1234567890"))

        with stack:
            await ux.check_for_updates(http_settings=http_settings, proxy_settings=proxy_settings)

        logger.warning.assert_not_called()
        logger.info.assert_not_called()

        json_loads.assert_called_once_with(_request.read.return_value)
        create_tcp_connector.assert_called_once_with(dns_cache=False, http_settings=http_settings)
        create_client_session.assert_called_once_with(
            connector=create_tcp_connector(),
            connector_owner=True,
            http_settings=http_settings,
            raise_for_status=True,
            trust_env=proxy_settings.trust_env,
        )
        _client_session.get.assert_called_once_with(
            "https://pypi.org/pypi/hikari/json",
            allow_redirects=http_settings.max_redirects is not None,
            max_redirects=http_settings.max_redirects,
            proxy=proxy_settings.url,
            proxy_headers=proxy_settings.all_headers,
        )

    @pytest.mark.parametrize("v", ["1.0.1", "1.0.1.dev10"])
    async def test_check_for_updates(self, v, http_settings, proxy_settings):
        data = {
            "releases": {
                v: [{"yanked": False}, {"yanked": True}],
                "1.0.0": [{"yanked": False}],
                "1.0.0.dev1": [{"yanked": False}],
                "0.1.0": [{"yanked": False}],
                "1.0.2": [{"yanked": True}],
            }
        }
        _request = hikari_test_helpers.AsyncContextManagerMock()
        _request.read = mock.AsyncMock()
        _client_session = hikari_test_helpers.AsyncContextManagerMock()
        _client_session.get = mock.Mock(return_value=_request)
        stack = contextlib.ExitStack()
        logger = stack.enter_context(mock.patch.object(ux, "_LOGGER"))
        json_loads = stack.enter_context(mock.patch.object(data_binding, "default_json_loads", return_value=data))
        create_client_session = stack.enter_context(
            mock.patch.object(net, "create_client_session", return_value=_client_session)
        )
        create_tcp_connector = stack.enter_context(mock.patch.object(net, "create_tcp_connector"))
        stack.enter_context(mock.patch.object(_about, "__version__", new="1.0.0.dev1"))
        stack.enter_context(mock.patch.object(_about, "__git_sha1__", new="1234567890"))

        with stack:
            await ux.check_for_updates(http_settings=http_settings, proxy_settings=proxy_settings)

        logger.warning.assert_not_called()
        logger.info.assert_called_once_with(
            "A newer version of hikari is available, consider upgrading to %s", ux.HikariVersion(v)
        )
        json_loads.assert_called_once_with(_request.read.return_value)
        create_tcp_connector.assert_called_once_with(dns_cache=False, http_settings=http_settings)
        create_client_session.assert_called_once_with(
            connector=create_tcp_connector(),
            connector_owner=True,
            http_settings=http_settings,
            raise_for_status=True,
            trust_env=proxy_settings.trust_env,
        )
        _client_session.get.assert_called_once_with(
            "https://pypi.org/pypi/hikari/json",
            allow_redirects=http_settings.max_redirects is not None,
            max_redirects=http_settings.max_redirects,
            proxy=proxy_settings.url,
            proxy_headers=proxy_settings.all_headers,
        )



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio

import mock.mock
import pytest

from hikari.internal import aio
from tests.hikari import hikari_test_helpers


class CoroutineStub:
    def __init__(self, *args, **kwargs):
        self.awaited = False
        self.args = args
        self.kwargs = kwargs

    def __eq__(self, other):
        return isinstance(other, CoroutineStub) and self.args == other.args and self.kwargs == other.kwargs

    def __await__(self):
        self.awaited = True
        # asyncio.sleep(0) is a special case that will just quickly return without doing anything special
        return asyncio.sleep(0)

    def __repr__(self):
        args = ", ".join(map(repr, self.args))
        kwargs = ", ".join(f"{key!s}={value!r}" for key, value in self.kwargs.items())
        return f"({args}, {kwargs})"


class CoroutineFunctionStub:
    def __call__(self, *args, **kwargs):
        return CoroutineStub(*args, **kwargs)


class TestCoroutineFunctionStubUsedInTests:
    def test_coro_stub_eq(self):
        assert CoroutineStub(9, 18, x=27) == CoroutineStub(9, 18, x=27)

    # POC for the stuff we use in tests
    def test_coro_stub_neq(self):
        assert CoroutineStub(9, 18, x=27) != CoroutineStub(9, 18, x=36)


class TestCompletedFuture:
    @pytest.mark.asyncio
    @pytest.mark.parametrize("args", [(), (12,)])
    async def test_is_awaitable(self, args):
        await aio.completed_future(*args)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("args", [(), (12,)])
    async def test_is_completed(self, args):
        future = aio.completed_future(*args)
        assert future.done()

    @pytest.mark.asyncio
    async def test_default_result_is_none(self):
        assert aio.completed_future().result() is None

    @pytest.mark.asyncio
    async def test_non_default_result(self):
        assert aio.completed_future(...).result() is ...


@pytest.mark.asyncio
class TestFirstCompleted:
    @hikari_test_helpers.timeout()
    async def test_first_future_completes(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        f1.set_result(None)
        await aio.first_completed(f1, f2, f3)
        assert f1.done()
        assert f2.cancelled()
        assert f3.cancelled()

    @hikari_test_helpers.timeout()
    async def test_second_future_completes(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        f2.set_result(None)
        await aio.first_completed(f1, f2, f3)
        assert f1.cancelled()
        assert f2.done()
        assert f3.cancelled()

    @hikari_test_helpers.timeout()
    async def test_timeout_propagates(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        with pytest.raises(asyncio.TimeoutError):
            await aio.first_completed(f1, f2, f3, timeout=0.01)

        assert f1.cancelled()
        assert f2.cancelled()
        assert f3.cancelled()

    @hikari_test_helpers.timeout()
    async def test_cancelled_propagates(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()
        f1.cancel()
        f2.cancel()
        f3.cancel()

        with pytest.raises(asyncio.CancelledError):
            await aio.first_completed(f1, f2, f3)

        assert f1.cancelled()
        assert f2.cancelled()
        assert f3.cancelled()

    @hikari_test_helpers.timeout()
    async def test_single_cancelled_propagates(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()
        f1.cancel()

        with pytest.raises(asyncio.CancelledError):
            await aio.first_completed(f1, f2, f3)

        assert f1.cancelled()
        assert f2.cancelled()
        assert f3.cancelled()

    # If the CI runners are slow, this may be flaky.
    @hikari_test_helpers.retry(3)
    @hikari_test_helpers.timeout()
    async def test_result_propagates(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()
        f3.set_result(22)

        await aio.first_completed(f1, f2, f3)

        assert f1.cancelled()
        assert f2.cancelled()
        assert f3.done()
        assert not f3.cancelled()

    # If the CI runners are slow, this may be flaky.
    @hikari_test_helpers.retry(3)
    @hikari_test_helpers.timeout()
    async def test_exception_propagates(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        error = FileExistsError("aaaa")
        f2.set_exception(error)

        with pytest.raises(FileExistsError):
            await aio.first_completed(f1, f2, f3)

        assert f1.cancelled()
        assert f2.done()
        assert f3.cancelled()


@pytest.mark.asyncio
class TestAllOf:
    # If the CI runners are slow, this may be flaky.
    @hikari_test_helpers.retry(3)
    @hikari_test_helpers.timeout()
    async def test_waits_for_all(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        async def quickly_run_task(task):
            try:
                await asyncio.wait_for(asyncio.shield(task), timeout=0.01)
            except asyncio.TimeoutError:
                pass

        waiter = event_loop.create_task(aio.all_of(f1, f2, f3))
        await quickly_run_task(waiter)
        assert not waiter.cancelled()
        assert not waiter.done()

        f1.set_result(1)
        await quickly_run_task(waiter)
        assert not waiter.cancelled()
        assert not waiter.done()

        f2.set_result(2)
        await quickly_run_task(waiter)
        assert not waiter.cancelled()
        assert not waiter.done()

        f3.set_result(3)
        await quickly_run_task(waiter)
        assert not waiter.cancelled()
        assert waiter.done()

    # If the CI runners are slow, this may be flaky.
    @hikari_test_helpers.retry(3)
    @hikari_test_helpers.timeout()
    async def test_cancels_all_if_one_errors(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        waiter = event_loop.create_task(aio.all_of(f1, f2, f3))
        assert not waiter.cancelled()
        assert not waiter.done()

        f2.set_exception(FileNotFoundError("boop"))
        with pytest.raises(FileNotFoundError, match="boop"):
            await waiter

        assert f1.cancelled()
        assert f3.cancelled()

    # If the CI runners are slow, this may be flaky.
    @hikari_test_helpers.retry(3)
    @hikari_test_helpers.timeout()
    async def test_cancels_all_if_timeout(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        waiter = event_loop.create_task(aio.all_of(f1, f2, f3, timeout=0.001))
        with pytest.raises(asyncio.TimeoutError):
            await waiter
        assert not waiter.cancelled(), "future was forcefully cancelled?"
        assert waiter.done(), "asyncio.TimeoutError not raised?"

    # If the CI runners are slow, this may be flaky.
    @hikari_test_helpers.retry(3)
    @hikari_test_helpers.timeout()
    async def test_cancels_all_if_cancelled(self):
        event_loop = asyncio.get_running_loop()

        f1 = event_loop.create_future()
        f2 = event_loop.create_future()
        f3 = event_loop.create_future()

        waiter = event_loop.create_task(aio.all_of(f1, f2, f3))
        event_loop.call_later(0.001, lambda *_: waiter.cancel())
        with pytest.raises(asyncio.CancelledError):
            await waiter

        assert waiter.cancelled(), "future was forcefully cancelled?"
        assert waiter.done(), "asyncio.CancelledError not raised?"


def test_get_or_make_loop():
    mock_loop = mock.Mock(asyncio.AbstractEventLoop, is_closed=mock.Mock(return_value=False))
    asyncio.set_event_loop(mock_loop)

    assert aio.get_or_make_loop() is mock_loop

    # Make sure to "cleanup" event loop or pytest_asyncio will error
    mock_loop.is_closed = mock.Mock(return_value=True)


def test_get_or_make_loop_handles_runtime_error():
    asyncio.set_event_loop(None)
    mock_loop = mock.Mock(asyncio.AbstractEventLoop)

    with mock.patch.object(asyncio, "new_event_loop", return_value=mock_loop) as new_event_loop:
        assert aio.get_or_make_loop() is mock_loop

        new_event_loop.assert_called_once_with()

    assert asyncio.get_event_loop_policy().get_event_loop() is mock_loop


def test_get_or_make_loop_handles_closed_loop():
    asyncio.set_event_loop(mock.Mock(asyncio.AbstractEventLoop, is_closed=mock.Mock(return_value=True)))
    mock_loop = mock.Mock(asyncio.AbstractEventLoop)

    with mock.patch.object(asyncio, "new_event_loop", return_value=mock_loop) as new_event_loop:
        assert aio.get_or_make_loop() is mock_loop

        new_event_loop.assert_called_once_with()

    assert asyncio.get_event_loop_policy().get_event_loop() is mock_loop
